from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FeeEventJSON(typing.TypedDict):
    index: int
    fees_collected_in_quote_lots: int


@dataclass
class FeeEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "index" / borsh.U16, "fees_collected_in_quote_lots" / borsh.U64
    )
    index: int
    fees_collected_in_quote_lots: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "FeeEvent":
        return cls(
            index=obj.index,
            fees_collected_in_quote_lots=obj.fees_collected_in_quote_lots,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "index": self.index,
            "fees_collected_in_quote_lots": self.fees_collected_in_quote_lots,
        }

    def to_json(self) -> FeeEventJSON:
        return {
            "index": self.index,
            "fees_collected_in_quote_lots": self.fees_collected_in_quote_lots,
        }

    @classmethod
    def from_json(cls, obj: FeeEventJSON) -> "FeeEvent":
        return cls(
            index=obj["index"],
            fees_collected_in_quote_lots=obj["fees_collected_in_quote_lots"],
        )



from __future__ import annotations
from . import (
    market_size_params,
    token_params,
)
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh

MARKET_HEADER_SIZE = 576


class MarketHeaderJSON(typing.TypedDict):
    discriminant: int
    status: int
    market_size_params: market_size_params.MarketSizeParamsJSON
    base_params: token_params.TokenParamsJSON
    base_lot_size: int
    quote_params: token_params.TokenParamsJSON
    quote_lot_size: int
    tick_size_in_quote_atoms_per_base_unit: int
    authority: str
    fee_recipient: str
    market_sequence_number: int
    successor: str
    raw_base_units_per_base_unit: int
    padding1: int
    padding2: list[int]


@dataclass
class MarketHeader:
    layout: typing.ClassVar = borsh.CStruct(
        "discriminant" / borsh.U64,
        "status" / borsh.U64,
        "market_size_params" / market_size_params.MarketSizeParams.layout,
        "base_params" / token_params.TokenParams.layout,
        "base_lot_size" / borsh.U64,
        "quote_params" / token_params.TokenParams.layout,
        "quote_lot_size" / borsh.U64,
        "tick_size_in_quote_atoms_per_base_unit" / borsh.U64,
        "authority" / BorshPubkey,
        "fee_recipient" / BorshPubkey,
        "market_sequence_number" / borsh.U64,
        "successor" / BorshPubkey,
        "raw_base_units_per_base_unit" / borsh.U32,
        "padding1" / borsh.U32,
        "padding2" / borsh.U64[32],
    )
    discriminant: int
    status: int
    market_size_params: market_size_params.MarketSizeParams
    base_params: token_params.TokenParams
    base_lot_size: int
    quote_params: token_params.TokenParams
    quote_lot_size: int
    tick_size_in_quote_atoms_per_base_unit: int
    authority: Pubkey
    fee_recipient: Pubkey
    market_sequence_number: int
    successor: Pubkey
    raw_base_units_per_base_unit: int
    padding1: int
    padding2: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "MarketHeader":
        return cls(
            discriminant=obj.discriminant,
            status=obj.status,
            market_size_params=market_size_params.MarketSizeParams.from_decoded(
                obj.market_size_params
            ),
            base_params=token_params.TokenParams.from_decoded(obj.base_params),
            base_lot_size=obj.base_lot_size,
            quote_params=token_params.TokenParams.from_decoded(obj.quote_params),
            quote_lot_size=obj.quote_lot_size,
            tick_size_in_quote_atoms_per_base_unit=obj.tick_size_in_quote_atoms_per_base_unit,
            authority=obj.authority,
            fee_recipient=obj.fee_recipient,
            market_sequence_number=obj.market_sequence_number,
            successor=obj.successor,
            raw_base_units_per_base_unit=obj.raw_base_units_per_base_unit,
            padding1=obj.padding1,
            padding2=obj.padding2,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "discriminant": self.discriminant,
            "status": self.status,
            "market_size_params": self.market_size_params.to_encodable(),
            "base_params": self.base_params.to_encodable(),
            "base_lot_size": self.base_lot_size,
            "quote_params": self.quote_params.to_encodable(),
            "quote_lot_size": self.quote_lot_size,
            "tick_size_in_quote_atoms_per_base_unit": self.tick_size_in_quote_atoms_per_base_unit,
            "authority": self.authority,
            "fee_recipient": self.fee_recipient,
            "market_sequence_number": self.market_sequence_number,
            "successor": self.successor,
            "raw_base_units_per_base_unit": self.raw_base_units_per_base_unit,
            "padding1": self.padding1,
            "padding2": self.padding2,
        }

    def to_json(self) -> MarketHeaderJSON:
        return {
            "discriminant": self.discriminant,
            "status": self.status,
            "market_size_params": self.market_size_params.to_json(),
            "base_params": self.base_params.to_json(),
            "base_lot_size": self.base_lot_size,
            "quote_params": self.quote_params.to_json(),
            "quote_lot_size": self.quote_lot_size,
            "tick_size_in_quote_atoms_per_base_unit": self.tick_size_in_quote_atoms_per_base_unit,
            "authority": str(self.authority),
            "fee_recipient": str(self.fee_recipient),
            "market_sequence_number": self.market_sequence_number,
            "successor": str(self.successor),
            "raw_base_units_per_base_unit": self.raw_base_units_per_base_unit,
            "padding1": self.padding1,
            "padding2": self.padding2,
        }

    @classmethod
    def from_json(cls, obj: MarketHeaderJSON) -> "MarketHeader":
        return cls(
            discriminant=obj["discriminant"],
            status=obj["status"],
            market_size_params=market_size_params.MarketSizeParams.from_json(
                obj["market_size_params"]
            ),
            base_params=token_params.TokenParams.from_json(obj["base_params"]),
            base_lot_size=obj["base_lot_size"],
            quote_params=token_params.TokenParams.from_json(obj["quote_params"]),
            quote_lot_size=obj["quote_lot_size"],
            tick_size_in_quote_atoms_per_base_unit=obj[
                "tick_size_in_quote_atoms_per_base_unit"
            ],
            authority=Pubkey.from_string(obj["authority"]),
            fee_recipient=Pubkey.from_string(obj["fee_recipient"]),
            market_sequence_number=obj["market_sequence_number"],
            successor=Pubkey.from_string(obj["successor"]),
            raw_base_units_per_base_unit=obj["raw_base_units_per_base_unit"],
            padding1=obj["padding1"],
            padding2=obj["padding2"],
        )



from typing import List, Optional, Tuple, Set, Any
import struct

from typing import List, Tuple, Dict

from solders.pubkey import Pubkey
from .market_metadata import MarketMetadata
from .types.market_header import MARKET_HEADER_SIZE, MarketHeader
from .types.fifo_order_id import FIFOOrderId
from .types.fifo_resting_order import FIFORestingOrder
from .types.trader_state import TraderState
from dataclasses import dataclass
from decimal import Decimal

DEFAULT_L2_LADDER_DEPTH = 10


@dataclass
class ActiveOrder:
    order_id: int
    market: Pubkey
    side: str
    price: Decimal
    price_in_ticks: int
    remaining_size: Decimal
    remaining_size_in_base_lots: int


class LadderLevel:
    price_in_ticks: Decimal
    size_in_base_lots: Decimal

    def __init__(self, price_in_ticks, size_in_base_lots):
        self.price_in_ticks = price_in_ticks
        self.size_in_base_lots = size_in_base_lots


class Ladder:
    bids: List[LadderLevel]
    asks: List[LadderLevel]

    def __init__(self):
        self.bids = []
        self.asks = []


class UiLadderLevel:
    price: Decimal
    size: Decimal

    def __init__(self, price, size):
        self.price = price
        self.size = size

    def __repr__(self) -> str:
        return f"{self.price} {self.size}"


class UiLadder:
    bids: List[UiLadderLevel]
    asks: List[UiLadderLevel]

    def __init__(self):
        self.bids = []
        self.asks = []

    def __repr__(self) -> str:
        ladder_str = ""
        for ask in reversed(self.asks):
            ladder_str += repr(ask)
            ladder_str += "\n"
        ladder_str += "\n"
        for bid in self.bids:
            ladder_str += repr(bid)
            ladder_str += "\n"
        return ladder_str


@dataclass
class Market:
    address: Pubkey
    metadata: MarketMetadata
    sequence_number: int
    base_lots_per_base_unit: int
    quote_lots_per_base_unit_per_tick: int
    order_sequence_number: int
    taker_fee_bps: int
    collected_quote_lot_fees: int
    unclaimed_quote_lot_fees: int
    bids: List[Tuple[FIFOOrderId, FIFORestingOrder]]
    asks: List[Tuple[FIFOOrderId, FIFORestingOrder]]
    traders: Dict[Pubkey, TraderState]
    trader_pubkey_to_trader_index: Dict[Pubkey, int]
    trader_index_to_trader_pubkey: Dict[int, Pubkey]

    @classmethod
    def deserialize_market_data(cls, market_pubkey: Pubkey, data: bytes) -> "Market":
        # Deserialize the market header

        header = MarketHeader.from_decoded(
            MarketHeader.layout.parse(data[:MARKET_HEADER_SIZE])
        )
        metadata = MarketMetadata(market_pubkey, header)
        offset = MARKET_HEADER_SIZE

        # Parse market data
        padding_len = 8 * 32
        remaining = data[offset + padding_len :]
        offset = 0

        base_lots_per_base_unit = int.from_bytes(
            remaining[offset : offset + 8], "little"
        )
        offset += 8
        quote_lots_per_base_unit_per_tick = int.from_bytes(
            remaining[offset : offset + 8], "little"
        )
        offset += 8
        sequence_number = int.from_bytes(remaining[offset : offset + 8], "little")
        offset += 8
        taker_fee_bps = int.from_bytes(remaining[offset : offset + 8], "little")
        offset += 8
        collected_quote_lot_fees = int.from_bytes(
            remaining[offset : offset + 8], "little"
        )
        offset += 8
        unclaimed_quote_lot_fees = int.from_bytes(
            remaining[offset : offset + 8], "little"
        )
        offset += 8
        remaining = remaining[offset:]

        # Parse bids, asks, and traders
        num_bids = header.market_size_params.bids_size
        num_asks = header.market_size_params.asks_size
        num_traders = header.market_size_params.num_seats

        bids_size = (
            16 + 16 + (16 + FIFOOrderId.size() + FIFORestingOrder.size()) * num_bids
        )
        asks_size = (
            16 + 16 + (16 + FIFOOrderId.size() + FIFORestingOrder.size()) * num_asks
        )
        traders_size = 16 + 16 + (16 + 32 + TraderState.size()) * num_traders
        offset = 0

        bid_buffer = remaining[offset : offset + bids_size]
        offset += bids_size
        ask_buffer = remaining[offset : offset + asks_size]
        offset += asks_size
        trader_buffer = remaining[offset : offset + traders_size]

        deserialize_fifo_order_id = lambda data, offset: FIFOOrderId.from_decoded(
            FIFOOrderId.layout.parse(data[offset : offset + FIFOOrderId.size()])
        )
        deserialize_fifo_resting_order = (
            lambda data, offset: FIFORestingOrder.from_decoded(
                FIFORestingOrder.layout.parse(
                    data[offset : offset + FIFORestingOrder.size()]
                )
            )
        )
        deserialize_pubkey = lambda data, offset: Pubkey(data[offset : offset + 32])
        deserialize_trader_state = lambda data, offset: TraderState.from_decoded(
            TraderState.layout.parse(data[offset : offset + TraderState.size()])
        )

        bids_unsorted = deserialize_red_black_tree(
            bid_buffer,
            FIFOOrderId.size(),
            FIFORestingOrder.size(),
            deserialize_fifo_order_id,
            deserialize_fifo_resting_order,
        )
        asks_unsorted = deserialize_red_black_tree(
            ask_buffer,
            FIFOOrderId.size(),
            FIFORestingOrder.size(),
            deserialize_fifo_order_id,
            deserialize_fifo_resting_order,
        )

        # Sort bids in descending order of price, and ascending order of sequence number
        bids = sorted(
            bids_unsorted,
            key=lambda x: (x[0].price_in_ticks, x[0].order_sequence_number),
            reverse=True,
        )

        # Sort asks in ascending order of price, and ascending order of sequence number
        asks = sorted(
            asks_unsorted,
            key=lambda x: (x[0].price_in_ticks, x[0].order_sequence_number),
        )

        traders = {}
        for k, trader_state in deserialize_red_black_tree(
            trader_buffer,
            32,
            TraderState.size(),
            deserialize_pubkey,
            deserialize_trader_state,
        ):
            traders[k] = trader_state

        trader_pubkey_to_trader_index = {}
        trader_index_to_trader_pubkey = {}
        for k, index in get_node_indices(
            trader_buffer,
            32,
            TraderState.size(),
            deserialize_pubkey,
            deserialize_trader_state,
        ).items():
            trader_pubkey_to_trader_index[k] = index
            trader_index_to_trader_pubkey[index] = k

        return cls(
            address=market_pubkey,
            metadata=metadata,
            sequence_number=header.market_sequence_number,
            base_lots_per_base_unit=base_lots_per_base_unit,
            quote_lots_per_base_unit_per_tick=quote_lots_per_base_unit_per_tick,
            order_sequence_number=sequence_number,
            taker_fee_bps=taker_fee_bps,
            collected_quote_lot_fees=collected_quote_lot_fees,
            unclaimed_quote_lot_fees=unclaimed_quote_lot_fees,
            bids=bids,
            asks=asks,
            traders=traders,
            trader_pubkey_to_trader_index=trader_pubkey_to_trader_index,
            trader_index_to_trader_pubkey=trader_index_to_trader_pubkey,
        )

    def get_bids(self, levels=1):
        return list(
            map(
                lambda order: (
                    self.metadata.ticks_to_float_price(order[0].price_in_ticks),
                    self.metadata.base_lots_to_raw_base_units_as_float(
                        order[1].num_base_lots
                    ),
                ),
                self.bids[:levels],
            )
        )

    def get_asks(self, levels=1):
        return list(
            map(
                lambda order: (
                    self.metadata.ticks_to_float_price(order[0].price_in_ticks),
                    self.metadata.base_lots_to_raw_base_units_as_float(
                        order[1].num_base_lots
                    ),
                ),
                self.asks[:levels],
            )
        )

    def get_orders_for_trader(
        self, trader: Pubkey
    ) -> Tuple[
        List[Tuple[FIFOOrderId, FIFORestingOrder]],  # bids
        List[Tuple[FIFOOrderId, FIFORestingOrder]],  # asks
    ]:
        trader_index = self.trader_pubkey_to_trader_index.get(trader)
        if trader_index is None:
            return ([], [])
        bid_orders = []
        for bid in self.bids:
            if bid[1].trader_index == trader_index:
                bid_orders.append(bid)
        ask_orders = []
        for ask in self.asks:
            if ask[1].trader_index == trader_index:
                ask_orders.append(ask)
        return (bid_orders, ask_orders)

    def get_active_orders(self, trader: Pubkey) -> List[ActiveOrder]:
        bid_orders, ask_orders = self.get_orders_for_trader(trader)
        active_orders = []
        for order_id, resting_order in bid_orders:
            active_orders.append(
                ActiveOrder(
                    order_id=order_id.to_int(),
                    market=self.address,
                    side="buy",
                    price=self.metadata.ticks_to_float_price(order_id.price_in_ticks),
                    price_in_ticks=order_id.price_in_ticks,
                    remaining_size=self.metadata.base_lots_to_raw_base_units_as_float(
                        resting_order.num_base_lots
                    ),
                    remaining_size_in_base_lots=resting_order.num_base_lots,
                )
            )
        for order_id, resting_order in ask_orders:
            active_orders.append(
                ActiveOrder(
                    order_id=order_id.to_int(),
                    market=self.address,
                    side="sell",
                    price=self.metadata.ticks_to_float_price(order_id.price_in_ticks),
                    price_in_ticks=order_id.price_in_ticks,
                    remaining_size=self.metadata.base_lots_to_raw_base_units_as_float(
                        resting_order.num_base_lots
                    ),
                    remaining_size_in_base_lots=resting_order.num_base_lots,
                )
            )
        return active_orders

    def get_trader_state(self, trader: Pubkey) -> Optional[TraderState]:
        return self.traders.get(trader)

    def get_ladder(
        self,
        slot: int = -1,
        unix_timestamp: int = -1,
        levels: int = DEFAULT_L2_LADDER_DEPTH,
    ) -> Ladder:
        ladder = Ladder()

        for order_id, resting_order in self.bids:
            if (
                resting_order.last_valid_slot != 0
                and resting_order.last_valid_slot < slot
            ):
                continue
            if (
                resting_order.last_valid_unix_timestamp_in_seconds != 0
                and resting_order.last_valid_unix_timestamp_in_seconds < unix_timestamp
            ):
                continue

            price_in_ticks = Decimal(order_id.price_in_ticks)
            size_in_base_lots = Decimal(resting_order.num_base_lots)
            if not ladder.bids:
                ladder.bids.append(LadderLevel(price_in_ticks, size_in_base_lots))
            else:
                prev = ladder.bids[-1]
                if price_in_ticks == prev.price_in_ticks:
                    prev.size_in_base_lots += size_in_base_lots
                else:
                    if len(ladder.bids) == levels:
                        break
                    ladder.bids.append(LadderLevel(price_in_ticks, size_in_base_lots))

        for order_id, resting_order in self.asks:
            if (
                resting_order.last_valid_slot != 0
                and resting_order.last_valid_slot < slot
            ):
                continue
            if (
                resting_order.last_valid_unix_timestamp_in_seconds != 0
                and resting_order.last_valid_unix_timestamp_in_seconds < unix_timestamp
            ):
                continue

            price_in_ticks = Decimal(order_id.price_in_ticks)
            size_in_base_lots = Decimal(resting_order.num_base_lots)
            if not ladder.asks:
                ladder.asks.append(LadderLevel(price_in_ticks, size_in_base_lots))
            else:
                prev = ladder.asks[-1]
                if price_in_ticks == prev.price_in_ticks:
                    prev.size_in_base_lots += size_in_base_lots
                else:
                    if len(ladder.asks) == levels:
                        break
                    ladder.asks.append(LadderLevel(price_in_ticks, size_in_base_lots))

        return ladder

    def get_ui_ladder(
        self,
        slot: int = -1,
        unix_timestamp: int = -1,
        levels: int = DEFAULT_L2_LADDER_DEPTH,
    ) -> UiLadder:
        ui_ladder = UiLadder()
        ladder = self.get_ladder(slot, unix_timestamp, levels)
        for bid in ladder.bids:
            ui_ladder.bids.append(
                UiLadderLevel(
                    Decimal(self.metadata.ticks_to_float_price(bid.price_in_ticks)),
                    Decimal(
                        self.metadata.base_lots_to_raw_base_units_as_float(
                            bid.size_in_base_lots
                        )
                    ),
                )
            )
        for ask in ladder.asks:
            ui_ladder.asks.append(
                UiLadderLevel(
                    Decimal(self.metadata.ticks_to_float_price(ask.price_in_ticks)),
                    Decimal(
                        self.metadata.base_lots_to_raw_base_units_as_float(
                            ask.size_in_base_lots
                        )
                    ),
                )
            )
        return ui_ladder


def deserialize_red_black_tree(
    data: bytes,
    key_size: int,
    value_size: int,
    deserialize_key,
    deserialize_value,
) -> list:
    tree_nodes = []
    nodes, free_nodes = deserialize_red_black_tree_nodes(
        data, key_size, value_size, deserialize_key, deserialize_value
    )

    for index, (key, value) in enumerate(nodes):
        if index not in free_nodes:
            tree_nodes.append((key, value))

    return tree_nodes


def get_node_indices(
    data: bytes,
    key_size: int,
    value_size: int,
    deserialize_key,
    deserialize_value,
) -> dict:
    index_map = {}
    nodes, free_nodes = deserialize_red_black_tree_nodes(
        data, key_size, value_size, deserialize_key, deserialize_value
    )

    for index, (key, _) in enumerate(nodes):
        if index not in free_nodes:
            index_map[key] = index + 1

    return index_map


def deserialize_red_black_tree_nodes(
    data: bytes,
    key_size: int,
    value_size: int,
    deserialize_key,
    deserialize_value,
) -> Tuple[List[Tuple[Any, Any]], Set[int]]:
    offset = 0

    nodes = []

    # Skip RBTree header
    offset += 16

    # Skip node allocator size
    offset += 8
    bump_index = struct.unpack_from("<i", data, offset)[0]
    offset += 4
    free_list_head = struct.unpack_from("<i", data, offset)[0]
    offset += 4

    free_list_pointers = []

    index = 0
    while offset < len(data) and index < bump_index - 1:
        registers = []
        for i in range(4):
            registers.append(struct.unpack_from("<i", data, offset)[0])
            offset += 4

        key = deserialize_key(data, offset)
        offset += key_size
        value = deserialize_value(data, offset)
        offset += value_size
        nodes.append((key, value))
        free_list_pointers.append((index, registers[0]))
        index += 1

    free_nodes = set()
    index_to_remove = free_list_head - 1

    counter = 0
    while free_list_head < bump_index:
        next_node = free_list_pointers[free_list_head - 1]
        index_to_remove, free_list_head = next_node
        free_nodes.add(index_to_remove)
        counter += 1
        if counter > bump_index:
            raise ValueError("Infinite loop detected")

    return nodes, free_nodes



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class TokenParamsJSON(typing.TypedDict):
    decimals: int
    vault_bump: int
    mint_key: str
    vault_key: str


@dataclass
class TokenParams:
    layout: typing.ClassVar = borsh.CStruct(
        "decimals" / borsh.U32,
        "vault_bump" / borsh.U32,
        "mint_key" / BorshPubkey,
        "vault_key" / BorshPubkey,
    )
    decimals: int
    vault_bump: int
    mint_key: Pubkey
    vault_key: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "TokenParams":
        return cls(
            decimals=obj.decimals,
            vault_bump=obj.vault_bump,
            mint_key=obj.mint_key,
            vault_key=obj.vault_key,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "decimals": self.decimals,
            "vault_bump": self.vault_bump,
            "mint_key": self.mint_key,
            "vault_key": self.vault_key,
        }

    def to_json(self) -> TokenParamsJSON:
        return {
            "decimals": self.decimals,
            "vault_bump": self.vault_bump,
            "mint_key": str(self.mint_key),
            "vault_key": str(self.vault_key),
        }

    @classmethod
    def from_json(cls, obj: TokenParamsJSON) -> "TokenParams":
        return cls(
            decimals=obj["decimals"],
            vault_bump=obj["vault_bump"],
            mint_key=Pubkey.from_string(obj["mint_key"]),
            vault_key=Pubkey.from_string(obj["vault_key"]),
        )



from __future__ import annotations
from . import (
    cancel_order_params,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class ReduceOrderParamsJSON(typing.TypedDict):
    base_params: cancel_order_params.CancelOrderParamsJSON
    size: int


@dataclass
class ReduceOrderParams:
    layout: typing.ClassVar = borsh.CStruct(
        "base_params" / cancel_order_params.CancelOrderParams.layout, "size" / borsh.U64
    )
    base_params: cancel_order_params.CancelOrderParams
    size: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "ReduceOrderParams":
        return cls(
            base_params=cancel_order_params.CancelOrderParams.from_decoded(
                obj.base_params
            ),
            size=obj.size,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"base_params": self.base_params.to_encodable(), "size": self.size}

    def to_json(self) -> ReduceOrderParamsJSON:
        return {"base_params": self.base_params.to_json(), "size": self.size}

    @classmethod
    def from_json(cls, obj: ReduceOrderParamsJSON) -> "ReduceOrderParams":
        return cls(
            base_params=cancel_order_params.CancelOrderParams.from_json(
                obj["base_params"]
            ),
            size=obj["size"],
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class CancelAllOrdersWithFreeFundsAccounts(typing.TypedDict):
    phoenix_program: Pubkey
    log_authority: Pubkey
    market: Pubkey
    trader: Pubkey


def cancel_all_orders_with_free_funds(
    accounts: CancelAllOrdersWithFreeFundsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["phoenix_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["log_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["trader"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x07"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
from . import (
    cancel_order_params,
)
import typing
from dataclasses import dataclass
from construct import Container, Construct
import borsh_construct as borsh


class CancelMultipleOrdersByIdParamsJSON(typing.TypedDict):
    orders: list[cancel_order_params.CancelOrderParamsJSON]


@dataclass
class CancelMultipleOrdersByIdParams:
    layout: typing.ClassVar = borsh.CStruct(
        "orders"
        / borsh.Vec(
            typing.cast(Construct, cancel_order_params.CancelOrderParams.layout)
        )
    )
    orders: list[cancel_order_params.CancelOrderParams]

    @classmethod
    def from_decoded(cls, obj: Container) -> "CancelMultipleOrdersByIdParams":
        return cls(
            orders=list(
                map(
                    lambda item: cancel_order_params.CancelOrderParams.from_decoded(
                        item
                    ),
                    obj.orders,
                )
            )
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"orders": list(map(lambda item: item.to_encodable(), self.orders))}

    def to_json(self) -> CancelMultipleOrdersByIdParamsJSON:
        return {"orders": list(map(lambda item: item.to_json(), self.orders))}

    @classmethod
    def from_json(
        cls, obj: CancelMultipleOrdersByIdParamsJSON
    ) -> "CancelMultipleOrdersByIdParams":
        return cls(
            orders=list(
                map(
                    lambda item: cancel_order_params.CancelOrderParams.from_json(item),
                    obj["orders"],
                )
            )
        )



from .swap import swap, SwapArgs, SwapAccounts
from .swap_with_free_funds import (
    swap_with_free_funds,
    SwapWithFreeFundsArgs,
    SwapWithFreeFundsAccounts,
)
from .place_limit_order import (
    place_limit_order,
    PlaceLimitOrderArgs,
    PlaceLimitOrderAccounts,
)
from .place_limit_order_with_free_funds import (
    place_limit_order_with_free_funds,
    PlaceLimitOrderWithFreeFundsArgs,
    PlaceLimitOrderWithFreeFundsAccounts,
)
from .reduce_order import reduce_order, ReduceOrderArgs, ReduceOrderAccounts
from .reduce_order_with_free_funds import (
    reduce_order_with_free_funds,
    ReduceOrderWithFreeFundsArgs,
    ReduceOrderWithFreeFundsAccounts,
)
from .cancel_all_orders import cancel_all_orders, CancelAllOrdersAccounts
from .cancel_all_orders_with_free_funds import (
    cancel_all_orders_with_free_funds,
    CancelAllOrdersWithFreeFundsAccounts,
)
from .cancel_up_to import cancel_up_to, CancelUpToArgs, CancelUpToAccounts
from .cancel_up_to_with_free_funds import (
    cancel_up_to_with_free_funds,
    CancelUpToWithFreeFundsArgs,
    CancelUpToWithFreeFundsAccounts,
)
from .cancel_multiple_orders_by_id import (
    cancel_multiple_orders_by_id,
    CancelMultipleOrdersByIdArgs,
    CancelMultipleOrdersByIdAccounts,
)
from .cancel_multiple_orders_by_id_with_free_funds import (
    cancel_multiple_orders_by_id_with_free_funds,
    CancelMultipleOrdersByIdWithFreeFundsArgs,
    CancelMultipleOrdersByIdWithFreeFundsAccounts,
)
from .withdraw_funds import withdraw_funds, WithdrawFundsArgs, WithdrawFundsAccounts
from .deposit_funds import deposit_funds, DepositFundsArgs, DepositFundsAccounts



from __future__ import annotations
from . import (
    condensed_order,
    failed_multiple_limit_order_behavior,
)
import typing
from dataclasses import dataclass
from construct import Container, Construct
import borsh_construct as borsh


class MultipleOrderPacketJSON(typing.TypedDict):
    bids: list[condensed_order.CondensedOrderJSON]
    asks: list[condensed_order.CondensedOrderJSON]
    client_order_id: typing.Optional[int]
    failed_multiple_limit_order_behavior: failed_multiple_limit_order_behavior.FailedMultipleLimitOrderBehaviorJSON


@dataclass
class MultipleOrderPacket:
    layout: typing.ClassVar = borsh.CStruct(
        "bids"
        / borsh.Vec(typing.cast(Construct, condensed_order.CondensedOrder.layout)),
        "asks"
        / borsh.Vec(typing.cast(Construct, condensed_order.CondensedOrder.layout)),
        "client_order_id" / borsh.Option(borsh.U128),
        "failed_multiple_limit_order_behavior"
        / failed_multiple_limit_order_behavior.layout,
    )
    bids: list[condensed_order.CondensedOrder]
    asks: list[condensed_order.CondensedOrder]
    client_order_id: typing.Optional[int]
    failed_multiple_limit_order_behavior: failed_multiple_limit_order_behavior.FailedMultipleLimitOrderBehaviorKind

    @classmethod
    def from_decoded(cls, obj: Container) -> "MultipleOrderPacket":
        return cls(
            bids=list(
                map(
                    lambda item: condensed_order.CondensedOrder.from_decoded(item),
                    obj.bids,
                )
            ),
            asks=list(
                map(
                    lambda item: condensed_order.CondensedOrder.from_decoded(item),
                    obj.asks,
                )
            ),
            client_order_id=obj.client_order_id,
            failed_multiple_limit_order_behavior=failed_multiple_limit_order_behavior.from_decoded(
                obj.failed_multiple_limit_order_behavior
            ),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "bids": list(map(lambda item: item.to_encodable(), self.bids)),
            "asks": list(map(lambda item: item.to_encodable(), self.asks)),
            "client_order_id": self.client_order_id,
            "failed_multiple_limit_order_behavior": self.failed_multiple_limit_order_behavior.to_encodable(),
        }

    def to_json(self) -> MultipleOrderPacketJSON:
        return {
            "bids": list(map(lambda item: item.to_json(), self.bids)),
            "asks": list(map(lambda item: item.to_json(), self.asks)),
            "client_order_id": self.client_order_id,
            "failed_multiple_limit_order_behavior": self.failed_multiple_limit_order_behavior.to_json(),
        }

    @classmethod
    def from_json(cls, obj: MultipleOrderPacketJSON) -> "MultipleOrderPacket":
        return cls(
            bids=list(
                map(
                    lambda item: condensed_order.CondensedOrder.from_json(item),
                    obj["bids"],
                )
            ),
            asks=list(
                map(
                    lambda item: condensed_order.CondensedOrder.from_json(item),
                    obj["asks"],
                )
            ),
            client_order_id=obj["client_order_id"],
            failed_multiple_limit_order_behavior=failed_multiple_limit_order_behavior.from_json(
                obj["failed_multiple_limit_order_behavior"]
            ),
        )



import typing
from . import market_size_params
from .market_size_params import MarketSizeParams, MarketSizeParamsJSON
from . import token_params
from .token_params import TokenParams, TokenParamsJSON
from . import seat
from .seat import Seat, SeatJSON
from . import audit_log_header
from .audit_log_header import AuditLogHeader, AuditLogHeaderJSON
from . import fill_event
from .fill_event import FillEvent, FillEventJSON
from . import reduce_event
from .reduce_event import ReduceEvent, ReduceEventJSON
from . import place_event
from .place_event import PlaceEvent, PlaceEventJSON
from . import evict_event
from .evict_event import EvictEvent, EvictEventJSON
from . import fill_summary_event
from .fill_summary_event import FillSummaryEvent, FillSummaryEventJSON
from . import fee_event
from .fee_event import FeeEvent, FeeEventJSON
from . import time_in_force_event
from .time_in_force_event import TimeInForceEvent, TimeInForceEventJSON
from . import expired_order_event
from .expired_order_event import ExpiredOrderEvent, ExpiredOrderEventJSON
from . import cancel_up_to_params
from .cancel_up_to_params import CancelUpToParams, CancelUpToParamsJSON
from . import cancel_multiple_orders_by_id_params
from .cancel_multiple_orders_by_id_params import (
    CancelMultipleOrdersByIdParams,
    CancelMultipleOrdersByIdParamsJSON,
)
from . import deposit_params
from .deposit_params import DepositParams, DepositParamsJSON
from . import multiple_order_packet
from .multiple_order_packet import MultipleOrderPacket, MultipleOrderPacketJSON
from . import condensed_order
from .condensed_order import CondensedOrder, CondensedOrderJSON
from . import cancel_order_params
from .cancel_order_params import CancelOrderParams, CancelOrderParamsJSON
from . import reduce_order_params
from .reduce_order_params import ReduceOrderParams, ReduceOrderParamsJSON
from . import withdraw_params
from .withdraw_params import WithdrawParams, WithdrawParamsJSON
from . import market_header
from .market_header import MarketHeader, MarketHeaderJSON, MARKET_HEADER_SIZE
from . import fifo_order_id
from .fifo_order_id import FIFOOrderId, FIFOOrderIdJSON
from . import phoenix_market_event
from .phoenix_market_event import PhoenixMarketEventKind, PhoenixMarketEventJSON
from . import failed_multiple_limit_order_behavior
from .failed_multiple_limit_order_behavior import (
    FailedMultipleLimitOrderBehaviorKind,
    FailedMultipleLimitOrderBehaviorJSON,
)
from . import market_status
from .market_status import MarketStatusKind, MarketStatusJSON
from . import seat_approval_status
from .seat_approval_status import SeatApprovalStatusKind, SeatApprovalStatusJSON
from . import order_packet
from .order_packet import OrderPacketKind, OrderPacketJSON
from . import side
from .side import SideKind, SideJSON
from . import self_trade_behavior
from .self_trade_behavior import SelfTradeBehaviorKind, SelfTradeBehaviorJSON



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TimeInForceEventJSON(typing.TypedDict):
    index: int
    order_sequence_number: int
    last_valid_slot: int
    last_valid_unix_timestamp_in_seconds: int


@dataclass
class TimeInForceEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "index" / borsh.U16,
        "order_sequence_number" / borsh.U64,
        "last_valid_slot" / borsh.U64,
        "last_valid_unix_timestamp_in_seconds" / borsh.U64,
    )
    index: int
    order_sequence_number: int
    last_valid_slot: int
    last_valid_unix_timestamp_in_seconds: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "TimeInForceEvent":
        return cls(
            index=obj.index,
            order_sequence_number=obj.order_sequence_number,
            last_valid_slot=obj.last_valid_slot,
            last_valid_unix_timestamp_in_seconds=obj.last_valid_unix_timestamp_in_seconds,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "index": self.index,
            "order_sequence_number": self.order_sequence_number,
            "last_valid_slot": self.last_valid_slot,
            "last_valid_unix_timestamp_in_seconds": self.last_valid_unix_timestamp_in_seconds,
        }

    def to_json(self) -> TimeInForceEventJSON:
        return {
            "index": self.index,
            "order_sequence_number": self.order_sequence_number,
            "last_valid_slot": self.last_valid_slot,
            "last_valid_unix_timestamp_in_seconds": self.last_valid_unix_timestamp_in_seconds,
        }

    @classmethod
    def from_json(cls, obj: TimeInForceEventJSON) -> "TimeInForceEvent":
        return cls(
            index=obj["index"],
            order_sequence_number=obj["order_sequence_number"],
            last_valid_slot=obj["last_valid_slot"],
            last_valid_unix_timestamp_in_seconds=obj[
                "last_valid_unix_timestamp_in_seconds"
            ],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class SeatJSON(typing.TypedDict):
    discriminant: int
    market: str
    trader: str
    approval_status: int
    padding: list[int]


@dataclass
class Seat:
    layout: typing.ClassVar = borsh.CStruct(
        "discriminant" / borsh.U64,
        "market" / BorshPubkey,
        "trader" / BorshPubkey,
        "approval_status" / borsh.U64,
        "padding" / borsh.U64[6],
    )
    discriminant: int
    market: Pubkey
    trader: Pubkey
    approval_status: int
    padding: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "Seat":
        return cls(
            discriminant=obj.discriminant,
            market=obj.market,
            trader=obj.trader,
            approval_status=obj.approval_status,
            padding=obj.padding,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "discriminant": self.discriminant,
            "market": self.market,
            "trader": self.trader,
            "approval_status": self.approval_status,
            "padding": self.padding,
        }

    def to_json(self) -> SeatJSON:
        return {
            "discriminant": self.discriminant,
            "market": str(self.market),
            "trader": str(self.trader),
            "approval_status": self.approval_status,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: SeatJSON) -> "Seat":
        return cls(
            discriminant=obj["discriminant"],
            market=Pubkey.from_string(obj["market"]),
            trader=Pubkey.from_string(obj["trader"]),
            approval_status=obj["approval_status"],
            padding=obj["padding"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class DepositParamsJSON(typing.TypedDict):
    quote_lots_to_deposit: int
    base_lots_to_deposit: int


@dataclass
class DepositParams:
    layout: typing.ClassVar = borsh.CStruct(
        "quote_lots_to_deposit" / borsh.U64, "base_lots_to_deposit" / borsh.U64
    )
    quote_lots_to_deposit: int
    base_lots_to_deposit: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "DepositParams":
        return cls(
            quote_lots_to_deposit=obj.quote_lots_to_deposit,
            base_lots_to_deposit=obj.base_lots_to_deposit,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "quote_lots_to_deposit": self.quote_lots_to_deposit,
            "base_lots_to_deposit": self.base_lots_to_deposit,
        }

    def to_json(self) -> DepositParamsJSON:
        return {
            "quote_lots_to_deposit": self.quote_lots_to_deposit,
            "base_lots_to_deposit": self.base_lots_to_deposit,
        }

    @classmethod
    def from_json(cls, obj: DepositParamsJSON) -> "DepositParams":
        return cls(
            quote_lots_to_deposit=obj["quote_lots_to_deposit"],
            base_lots_to_deposit=obj["base_lots_to_deposit"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class ActiveJSON(typing.TypedDict):
    kind: typing.Literal["Active"]


class PostOnlyJSON(typing.TypedDict):
    kind: typing.Literal["PostOnly"]


class PausedJSON(typing.TypedDict):
    kind: typing.Literal["Paused"]


class ClosedJSON(typing.TypedDict):
    kind: typing.Literal["Closed"]


class TombstonedJSON(typing.TypedDict):
    kind: typing.Literal["Tombstoned"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Active:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Active"

    @classmethod
    def to_json(cls) -> ActiveJSON:
        return ActiveJSON(
            kind="Active",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Active": {},
        }


@dataclass
class PostOnly:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "PostOnly"

    @classmethod
    def to_json(cls) -> PostOnlyJSON:
        return PostOnlyJSON(
            kind="PostOnly",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PostOnly": {},
        }


@dataclass
class Paused:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Paused"

    @classmethod
    def to_json(cls) -> PausedJSON:
        return PausedJSON(
            kind="Paused",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Paused": {},
        }


@dataclass
class Closed:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "Closed"

    @classmethod
    def to_json(cls) -> ClosedJSON:
        return ClosedJSON(
            kind="Closed",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Closed": {},
        }


@dataclass
class Tombstoned:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "Tombstoned"

    @classmethod
    def to_json(cls) -> TombstonedJSON:
        return TombstonedJSON(
            kind="Tombstoned",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Tombstoned": {},
        }


MarketStatusKind = typing.Union[
    Uninitialized, Active, PostOnly, Paused, Closed, Tombstoned
]
MarketStatusJSON = typing.Union[
    UninitializedJSON, ActiveJSON, PostOnlyJSON, PausedJSON, ClosedJSON, TombstonedJSON
]


def from_decoded(obj: dict) -> MarketStatusKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Active" in obj:
        return Active()
    if "PostOnly" in obj:
        return PostOnly()
    if "Paused" in obj:
        return Paused()
    if "Closed" in obj:
        return Closed()
    if "Tombstoned" in obj:
        return Tombstoned()
    raise ValueError("Invalid enum object")


def from_json(obj: MarketStatusJSON) -> MarketStatusKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Active":
        return Active()
    if obj["kind"] == "PostOnly":
        return PostOnly()
    if obj["kind"] == "Paused":
        return Paused()
    if obj["kind"] == "Closed":
        return Closed()
    if obj["kind"] == "Tombstoned":
        return Tombstoned()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "Active" / borsh.CStruct(),
    "PostOnly" / borsh.CStruct(),
    "Paused" / borsh.CStruct(),
    "Closed" / borsh.CStruct(),
    "Tombstoned" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FillSummaryEventJSON(typing.TypedDict):
    index: int
    client_order_id: int
    total_base_lots_filled: int
    total_quote_lots_filled: int
    total_fee_in_quote_lots: int


@dataclass
class FillSummaryEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "index" / borsh.U16,
        "client_order_id" / borsh.U128,
        "total_base_lots_filled" / borsh.U64,
        "total_quote_lots_filled" / borsh.U64,
        "total_fee_in_quote_lots" / borsh.U64,
    )
    index: int
    client_order_id: int
    total_base_lots_filled: int
    total_quote_lots_filled: int
    total_fee_in_quote_lots: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "FillSummaryEvent":
        return cls(
            index=obj.index,
            client_order_id=obj.client_order_id,
            total_base_lots_filled=obj.total_base_lots_filled,
            total_quote_lots_filled=obj.total_quote_lots_filled,
            total_fee_in_quote_lots=obj.total_fee_in_quote_lots,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "index": self.index,
            "client_order_id": self.client_order_id,
            "total_base_lots_filled": self.total_base_lots_filled,
            "total_quote_lots_filled": self.total_quote_lots_filled,
            "total_fee_in_quote_lots": self.total_fee_in_quote_lots,
        }

    def to_json(self) -> FillSummaryEventJSON:
        return {
            "index": self.index,
            "client_order_id": self.client_order_id,
            "total_base_lots_filled": self.total_base_lots_filled,
            "total_quote_lots_filled": self.total_quote_lots_filled,
            "total_fee_in_quote_lots": self.total_fee_in_quote_lots,
        }

    @classmethod
    def from_json(cls, obj: FillSummaryEventJSON) -> "FillSummaryEvent":
        return cls(
            index=obj["index"],
            client_order_id=obj["client_order_id"],
            total_base_lots_filled=obj["total_base_lots_filled"],
            total_quote_lots_filled=obj["total_quote_lots_filled"],
            total_fee_in_quote_lots=obj["total_fee_in_quote_lots"],
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class CancelUpToWithFreeFundsArgs(typing.TypedDict):
    params: types.cancel_up_to_params.CancelUpToParams


layout = borsh.CStruct("params" / types.cancel_up_to_params.CancelUpToParams.layout)


class CancelUpToWithFreeFundsAccounts(typing.TypedDict):
    phoenix_program: Pubkey
    log_authority: Pubkey
    market: Pubkey
    trader: Pubkey


def cancel_up_to_with_free_funds(
    args: CancelUpToWithFreeFundsArgs,
    accounts: CancelUpToWithFreeFundsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["phoenix_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["log_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["trader"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x09"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FIFOOrderIdJSON(typing.TypedDict):
    price_in_ticks: int
    order_sequence_number: int


@dataclass
class FIFOOrderId:
    layout: typing.ClassVar = borsh.CStruct(
        "price_in_ticks" / borsh.U64, "order_sequence_number" / borsh.U64
    )
    price_in_ticks: int
    order_sequence_number: int

    @staticmethod
    def size():
        return 16

    @classmethod
    def from_decoded(cls, obj: Container) -> "FIFOOrderId":
        return cls(
            price_in_ticks=obj.price_in_ticks,
            order_sequence_number=obj.order_sequence_number,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "price_in_ticks": self.price_in_ticks,
            "order_sequence_number": self.order_sequence_number,
        }

    def to_json(self) -> FIFOOrderIdJSON:
        return {
            "price_in_ticks": self.price_in_ticks,
            "order_sequence_number": self.order_sequence_number,
        }

    @classmethod
    def from_json(cls, obj: FIFOOrderIdJSON) -> "FIFOOrderId":
        return cls(
            price_in_ticks=obj["price_in_ticks"],
            order_sequence_number=obj["order_sequence_number"],
        )

    def __hash__(self):
        return hash((self.price_in_ticks, self.order_sequence_number))

    def __repr__(self) -> str:
        return f"FIFOOrderId(price_in_ticks={self.price_in_ticks}, order_sequence_number={self.order_sequence_number})"

    def __lt__(self, other: FIFOOrderId) -> bool:
        if self.order_sequence_number & 1 << 63 != 0:
            if self.price_in_ticks == other.price_in_ticks:
                return self.order_sequence_number < other.order_sequence_number
            else:
                return self.price_in_ticks < other.price_in_ticks
        else:
            if self.price_in_ticks == other.price_in_ticks:
                return self.order_sequence_number > other.order_sequence_number
            else:
                return self.price_in_ticks > other.price_in_ticks

    def to_int(self) -> int:
        return self.price_in_ticks << 64 | self.order_sequence_number

    @classmethod
    def from_int(cls, order_id) -> "FIFOOrderId":
        return FIFOOrderId(
            price_in_ticks=order_id >> 64,
            order_sequence_number=order_id & 0xFFFFFFFFFFFFFFFF,
        )



from dataclasses import dataclass
from decimal import Decimal
from traceback import StackSummary
from typing import Optional, Union

from phoenix.types.fifo_order_id import FIFOOrderId
from phoenix.types.side import Ask, Bid
from solders.pubkey import Pubkey


@dataclass
class FilledOrder:
    order_id: int
    market_pubkey: Pubkey
    sequence_number: int
    slot: int
    unix_timestamp_in_seconds: int
    taker_side: Union[Bid, Ask]
    price: Decimal
    quantity_filled: Decimal
    quantity_remaining: Decimal
    taker_pubkey: Pubkey
    maker_pubkey: Pubkey


@dataclass
class OpenOrder:
    order_id: int
    client_order_id: int
    market_pubkey: Pubkey
    sequence_number: int
    slot: int
    side: Union[Bid, Ask]
    price: Decimal
    quantity_placed: Decimal
    unix_timestamp_in_seconds: int
    side: Union[Bid, Ask]
    maker_pubkey: Pubkey


@dataclass
class CancelledOrder:
    order_id: int
    market_pubkey: Pubkey
    sequence_number: int
    slot: int
    unix_timestamp_in_seconds: int
    side: Union[Bid, Ask]
    price: Decimal
    quantity_removed: Decimal
    quantity_remaining: Decimal
    maker_pubkey: Pubkey


@dataclass
class OrderSubscribeError:
    message: str
    reconnection_count: int
    reconnected: bool
    error_count: int
    traceback: Optional[StackSummary]
    exception: Exception


OrderSubscribeResponse = Union[
    FilledOrder,
    OpenOrder,
    CancelledOrder,
    OrderSubscribeError,
]



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class NotApprovedJSON(typing.TypedDict):
    kind: typing.Literal["NotApproved"]


class ApprovedJSON(typing.TypedDict):
    kind: typing.Literal["Approved"]


class RetiredJSON(typing.TypedDict):
    kind: typing.Literal["Retired"]


@dataclass
class NotApproved:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "NotApproved"

    @classmethod
    def to_json(cls) -> NotApprovedJSON:
        return NotApprovedJSON(
            kind="NotApproved",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "NotApproved": {},
        }


@dataclass
class Approved:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Approved"

    @classmethod
    def to_json(cls) -> ApprovedJSON:
        return ApprovedJSON(
            kind="Approved",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Approved": {},
        }


@dataclass
class Retired:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Retired"

    @classmethod
    def to_json(cls) -> RetiredJSON:
        return RetiredJSON(
            kind="Retired",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Retired": {},
        }


SeatApprovalStatusKind = typing.Union[NotApproved, Approved, Retired]
SeatApprovalStatusJSON = typing.Union[NotApprovedJSON, ApprovedJSON, RetiredJSON]


def from_decoded(obj: dict) -> SeatApprovalStatusKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "NotApproved" in obj:
        return NotApproved()
    if "Approved" in obj:
        return Approved()
    if "Retired" in obj:
        return Retired()
    raise ValueError("Invalid enum object")


def from_json(obj: SeatApprovalStatusJSON) -> SeatApprovalStatusKind:
    if obj["kind"] == "NotApproved":
        return NotApproved()
    if obj["kind"] == "Approved":
        return Approved()
    if obj["kind"] == "Retired":
        return Retired()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "NotApproved" / borsh.CStruct(),
    "Approved" / borsh.CStruct(),
    "Retired" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class AbortJSON(typing.TypedDict):
    kind: typing.Literal["Abort"]


class CancelProvideJSON(typing.TypedDict):
    kind: typing.Literal["CancelProvide"]


class DecrementTakeJSON(typing.TypedDict):
    kind: typing.Literal["DecrementTake"]


@dataclass
class Abort:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Abort"

    @classmethod
    def to_json(cls) -> AbortJSON:
        return AbortJSON(
            kind="Abort",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Abort": {},
        }


@dataclass
class CancelProvide:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "CancelProvide"

    @classmethod
    def to_json(cls) -> CancelProvideJSON:
        return CancelProvideJSON(
            kind="CancelProvide",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "CancelProvide": {},
        }


@dataclass
class DecrementTake:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "DecrementTake"

    @classmethod
    def to_json(cls) -> DecrementTakeJSON:
        return DecrementTakeJSON(
            kind="DecrementTake",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "DecrementTake": {},
        }


SelfTradeBehaviorKind = typing.Union[Abort, CancelProvide, DecrementTake]
SelfTradeBehaviorJSON = typing.Union[AbortJSON, CancelProvideJSON, DecrementTakeJSON]


def from_decoded(obj: dict) -> SelfTradeBehaviorKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Abort" in obj:
        return Abort()
    if "CancelProvide" in obj:
        return CancelProvide()
    if "DecrementTake" in obj:
        return DecrementTake()
    raise ValueError("Invalid enum object")


def from_json(obj: SelfTradeBehaviorJSON) -> SelfTradeBehaviorKind:
    if obj["kind"] == "Abort":
        return Abort()
    if obj["kind"] == "CancelProvide":
        return CancelProvide()
    if obj["kind"] == "DecrementTake":
        return DecrementTake()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Abort" / borsh.CStruct(),
    "CancelProvide" / borsh.CStruct(),
    "DecrementTake" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TraderStateJSON(typing.TypedDict):
    quote_lots_locked: int
    quote_lots_free: int
    base_lots_locked: int
    base_lots_free: int
    padding: list[int]


@dataclass
class TraderState:
    layout: typing.ClassVar = borsh.CStruct(
        "quote_lots_locked" / borsh.U64,
        "quote_lots_free" / borsh.U64,
        "base_lots_locked" / borsh.U64,
        "base_lots_free" / borsh.U64,
        "padding" / borsh.U64[8],
    )
    quote_lots_locked: int
    quote_lots_free: int
    base_lots_locked: int
    base_lots_free: int
    padding: list[int]

    @staticmethod
    def size():
        return 96

    @classmethod
    def from_decoded(cls, obj: Container) -> "TraderState":
        return cls(
            quote_lots_locked=obj.quote_lots_locked,
            quote_lots_free=obj.quote_lots_free,
            base_lots_locked=obj.base_lots_locked,
            base_lots_free=obj.base_lots_free,
            padding=obj.padding,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "quote_lots_locked": self.quote_lots_locked,
            "quote_lots_free": self.quote_lots_free,
            "base_lots_locked": self.base_lots_locked,
            "base_lots_free": self.base_lots_free,
            "padding": self.padding,
        }

    def to_json(self) -> TraderStateJSON:
        return {
            "quote_lots_locked": self.quote_lots_locked,
            "quote_lots_free": self.quote_lots_free,
            "base_lots_locked": self.base_lots_locked,
            "base_lots_free": self.base_lots_free,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: TraderStateJSON) -> "TraderState":
        return cls(
            quote_lots_locked=obj["quote_lots_locked"],
            quote_lots_free=obj["quote_lots_free"],
            base_lots_locked=obj["base_lots_locked"],
            base_lots_free=obj["base_lots_free"],
            padding=obj["padding"],
        )

    def __eq__(self, other: TraderState) -> bool:
        return (
            self.quote_lots_locked == other.quote_lots_locked
            and self.quote_lots_free == other.quote_lots_free
            and self.base_lots_locked == other.base_lots_locked
            and self.base_lots_free == other.base_lots_free
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class MarketSizeParamsJSON(typing.TypedDict):
    bids_size: int
    asks_size: int
    num_seats: int


@dataclass
class MarketSizeParams:
    layout: typing.ClassVar = borsh.CStruct(
        "bids_size" / borsh.U64, "asks_size" / borsh.U64, "num_seats" / borsh.U64
    )
    bids_size: int
    asks_size: int
    num_seats: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "MarketSizeParams":
        return cls(
            bids_size=obj.bids_size, asks_size=obj.asks_size, num_seats=obj.num_seats
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "bids_size": self.bids_size,
            "asks_size": self.asks_size,
            "num_seats": self.num_seats,
        }

    def to_json(self) -> MarketSizeParamsJSON:
        return {
            "bids_size": self.bids_size,
            "asks_size": self.asks_size,
            "num_seats": self.num_seats,
        }

    @classmethod
    def from_json(cls, obj: MarketSizeParamsJSON) -> "MarketSizeParams":
        return cls(
            bids_size=obj["bids_size"],
            asks_size=obj["asks_size"],
            num_seats=obj["num_seats"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class FailOnInsufficientFundsAndAmendOnCrossJSON(typing.TypedDict):
    kind: typing.Literal["FailOnInsufficientFundsAndAmendOnCross"]


class FailOnInsufficientFundsAndFailOnCrossJSON(typing.TypedDict):
    kind: typing.Literal["FailOnInsufficientFundsAndFailOnCross"]


class SkipOnInsufficientFundsAndAmendOnCrossJSON(typing.TypedDict):
    kind: typing.Literal["SkipOnInsufficientFundsAndAmendOnCross"]


class SkipOnInsufficientFundsAndFailOnCrossJSON(typing.TypedDict):
    kind: typing.Literal["SkipOnInsufficientFundsAndFailOnCross"]


@dataclass
class FailOnInsufficientFundsAndAmendOnCross:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "FailOnInsufficientFundsAndAmendOnCross"

    @classmethod
    def to_json(cls) -> FailOnInsufficientFundsAndAmendOnCrossJSON:
        return FailOnInsufficientFundsAndAmendOnCrossJSON(
            kind="FailOnInsufficientFundsAndAmendOnCross",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FailOnInsufficientFundsAndAmendOnCross": {},
        }


@dataclass
class FailOnInsufficientFundsAndFailOnCross:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "FailOnInsufficientFundsAndFailOnCross"

    @classmethod
    def to_json(cls) -> FailOnInsufficientFundsAndFailOnCrossJSON:
        return FailOnInsufficientFundsAndFailOnCrossJSON(
            kind="FailOnInsufficientFundsAndFailOnCross",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FailOnInsufficientFundsAndFailOnCross": {},
        }


@dataclass
class SkipOnInsufficientFundsAndAmendOnCross:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "SkipOnInsufficientFundsAndAmendOnCross"

    @classmethod
    def to_json(cls) -> SkipOnInsufficientFundsAndAmendOnCrossJSON:
        return SkipOnInsufficientFundsAndAmendOnCrossJSON(
            kind="SkipOnInsufficientFundsAndAmendOnCross",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SkipOnInsufficientFundsAndAmendOnCross": {},
        }


@dataclass
class SkipOnInsufficientFundsAndFailOnCross:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "SkipOnInsufficientFundsAndFailOnCross"

    @classmethod
    def to_json(cls) -> SkipOnInsufficientFundsAndFailOnCrossJSON:
        return SkipOnInsufficientFundsAndFailOnCrossJSON(
            kind="SkipOnInsufficientFundsAndFailOnCross",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "SkipOnInsufficientFundsAndFailOnCross": {},
        }


FailedMultipleLimitOrderBehaviorKind = typing.Union[
    FailOnInsufficientFundsAndAmendOnCross,
    FailOnInsufficientFundsAndFailOnCross,
    SkipOnInsufficientFundsAndAmendOnCross,
    SkipOnInsufficientFundsAndFailOnCross,
]
FailedMultipleLimitOrderBehaviorJSON = typing.Union[
    FailOnInsufficientFundsAndAmendOnCrossJSON,
    FailOnInsufficientFundsAndFailOnCrossJSON,
    SkipOnInsufficientFundsAndAmendOnCrossJSON,
    SkipOnInsufficientFundsAndFailOnCrossJSON,
]


def from_decoded(obj: dict) -> FailedMultipleLimitOrderBehaviorKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "FailOnInsufficientFundsAndAmendOnCross" in obj:
        return FailOnInsufficientFundsAndAmendOnCross()
    if "FailOnInsufficientFundsAndFailOnCross" in obj:
        return FailOnInsufficientFundsAndFailOnCross()
    if "SkipOnInsufficientFundsAndAmendOnCross" in obj:
        return SkipOnInsufficientFundsAndAmendOnCross()
    if "SkipOnInsufficientFundsAndFailOnCross" in obj:
        return SkipOnInsufficientFundsAndFailOnCross()
    raise ValueError("Invalid enum object")


def from_json(
    obj: FailedMultipleLimitOrderBehaviorJSON,
) -> FailedMultipleLimitOrderBehaviorKind:
    if obj["kind"] == "FailOnInsufficientFundsAndAmendOnCross":
        return FailOnInsufficientFundsAndAmendOnCross()
    if obj["kind"] == "FailOnInsufficientFundsAndFailOnCross":
        return FailOnInsufficientFundsAndFailOnCross()
    if obj["kind"] == "SkipOnInsufficientFundsAndAmendOnCross":
        return SkipOnInsufficientFundsAndAmendOnCross()
    if obj["kind"] == "SkipOnInsufficientFundsAndFailOnCross":
        return SkipOnInsufficientFundsAndFailOnCross()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "FailOnInsufficientFundsAndAmendOnCross" / borsh.CStruct(),
    "FailOnInsufficientFundsAndFailOnCross" / borsh.CStruct(),
    "SkipOnInsufficientFundsAndAmendOnCross" / borsh.CStruct(),
    "SkipOnInsufficientFundsAndFailOnCross" / borsh.CStruct(),
)



from __future__ import annotations
from . import (
    side,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class CancelUpToParamsJSON(typing.TypedDict):
    side: side.SideJSON
    tick_limit: typing.Optional[int]
    num_orders_to_search: typing.Optional[int]
    num_orders_to_cancel: typing.Optional[int]


@dataclass
class CancelUpToParams:
    layout: typing.ClassVar = borsh.CStruct(
        "side" / side.layout,
        "tick_limit" / borsh.Option(borsh.U64),
        "num_orders_to_search" / borsh.Option(borsh.U32),
        "num_orders_to_cancel" / borsh.Option(borsh.U32),
    )
    side: side.SideKind
    tick_limit: typing.Optional[int]
    num_orders_to_search: typing.Optional[int]
    num_orders_to_cancel: typing.Optional[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "CancelUpToParams":
        return cls(
            side=side.from_decoded(obj.side),
            tick_limit=obj.tick_limit,
            num_orders_to_search=obj.num_orders_to_search,
            num_orders_to_cancel=obj.num_orders_to_cancel,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "side": self.side.to_encodable(),
            "tick_limit": self.tick_limit,
            "num_orders_to_search": self.num_orders_to_search,
            "num_orders_to_cancel": self.num_orders_to_cancel,
        }

    def to_json(self) -> CancelUpToParamsJSON:
        return {
            "side": self.side.to_json(),
            "tick_limit": self.tick_limit,
            "num_orders_to_search": self.num_orders_to_search,
            "num_orders_to_cancel": self.num_orders_to_cancel,
        }

    @classmethod
    def from_json(cls, obj: CancelUpToParamsJSON) -> "CancelUpToParams":
        return cls(
            side=side.from_json(obj["side"]),
            tick_limit=obj["tick_limit"],
            num_orders_to_search=obj["num_orders_to_search"],
            num_orders_to_cancel=obj["num_orders_to_cancel"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class WithdrawParamsJSON(typing.TypedDict):
    quote_lots_to_withdraw: typing.Optional[int]
    base_lots_to_withdraw: typing.Optional[int]


@dataclass
class WithdrawParams:
    layout: typing.ClassVar = borsh.CStruct(
        "quote_lots_to_withdraw" / borsh.Option(borsh.U64),
        "base_lots_to_withdraw" / borsh.Option(borsh.U64),
    )
    quote_lots_to_withdraw: typing.Optional[int]
    base_lots_to_withdraw: typing.Optional[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "WithdrawParams":
        return cls(
            quote_lots_to_withdraw=obj.quote_lots_to_withdraw,
            base_lots_to_withdraw=obj.base_lots_to_withdraw,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "quote_lots_to_withdraw": self.quote_lots_to_withdraw,
            "base_lots_to_withdraw": self.base_lots_to_withdraw,
        }

    def to_json(self) -> WithdrawParamsJSON:
        return {
            "quote_lots_to_withdraw": self.quote_lots_to_withdraw,
            "base_lots_to_withdraw": self.base_lots_to_withdraw,
        }

    @classmethod
    def from_json(cls, obj: WithdrawParamsJSON) -> "WithdrawParams":
        return cls(
            quote_lots_to_withdraw=obj["quote_lots_to_withdraw"],
            base_lots_to_withdraw=obj["base_lots_to_withdraw"],
        )



from typing import Union
from phoenix.instructions import swap
from phoenix.instructions.place_limit_order import (
    PlaceLimitOrderAccounts,
    PlaceLimitOrderArgs,
    place_limit_order,
)
from phoenix.instructions.swap import SwapAccounts, SwapArgs
from phoenix.program_id import PROGRAM_ID
from phoenix.types.market_status import PostOnly
from phoenix.types.order_packet import ImmediateOrCancel, Limit
from .types.market_header import MarketHeader
from solders.pubkey import Pubkey
from spl.token.instructions import get_associated_token_address
from solana.transaction import Instruction


class MarketMetadata:
    def __init__(self, market_pubkey: Pubkey, header: MarketHeader):
        self.address = market_pubkey
        self.base_mint = header.base_params.mint_key
        self.base_vault = header.base_params.vault_key
        self.quote_mint = header.quote_params.mint_key
        self.quote_vault = header.quote_params.vault_key
        self.base_decimals = header.base_params.decimals
        self.quote_decimals = header.quote_params.decimals
        self.base_atoms_per_raw_base_unit = 10**self.base_decimals
        self.quote_atoms_per_quote_unit = 10**self.quote_decimals
        self.quote_atoms_per_quote_lot = header.quote_lot_size
        self.base_atoms_per_base_lot = header.base_lot_size
        self.tick_size_in_quote_atoms_per_base_unit = (
            header.tick_size_in_quote_atoms_per_base_unit
        )
        self.raw_base_units_per_base_unit = max(header.raw_base_units_per_base_unit, 1)

        if (
            self.base_atoms_per_raw_base_unit
            * self.raw_base_units_per_base_unit
            % self.base_atoms_per_base_lot
            != 0
        ):
            raise ValueError("Invalid base lot size (in base atoms per base lot)")

        self.num_base_lots_per_base_unit = (
            self.base_atoms_per_raw_base_unit * self.raw_base_units_per_base_unit
        ) // self.base_atoms_per_base_lot
        self.market_size_params = header.market_size_params

    def raw_base_units_to_base_lots_rounded_down(self, raw_base_units):
        base_units = raw_base_units / self.raw_base_units_per_base_unit
        return int(base_units * self.num_base_lots_per_base_unit)

    def raw_base_units_to_base_lots_rounded_up(self, raw_base_units):
        base_units = raw_base_units / self.raw_base_units_per_base_unit
        return int(
            (base_units * self.num_base_lots_per_base_unit) + 0.5
        )  # using 0.5 to achieve rounding up

    def base_atoms_to_base_lots_rounded_down(self, base_atoms):
        return base_atoms // self.base_atoms_per_base_lot

    def base_atoms_to_base_lots_rounded_up(self, base_atoms):
        return (
            base_atoms + self.base_atoms_per_base_lot - 1
        ) // self.base_atoms_per_base_lot

    def base_lots_to_base_atoms(self, base_lots):
        return base_lots * self.base_atoms_per_base_lot

    def quote_units_to_quote_lots(self, quote_units):
        return int(
            quote_units
            * (self.quote_atoms_per_quote_unit / self.quote_atoms_per_quote_lot)
        )

    def quote_atoms_to_quote_lots_rounded_down(self, quote_atoms):
        return quote_atoms // self.quote_atoms_per_quote_lot

    def quote_atoms_to_quote_lots_rounded_up(self, quote_atoms):
        return (
            quote_atoms + self.quote_atoms_per_quote_lot - 1
        ) // self.quote_atoms_per_quote_lot

    def quote_lots_to_quote_atoms(self, quote_lots):
        return quote_lots * self.quote_atoms_per_quote_lot

    def quote_lots_to_quote_units(self, quote_lots):
        return quote_lots * (
            self.quote_atoms_per_quote_lot / self.quote_atoms_per_quote_unit
        )

    def base_atoms_to_raw_base_units_as_float(self, base_atoms):
        return base_atoms / self.base_atoms_per_raw_base_unit

    def base_lots_to_raw_base_units_as_float(self, base_lots):
        return self.base_atoms_to_raw_base_units_as_float(
            self.base_lots_to_base_atoms(base_lots)
        )

    def quote_atoms_to_quote_units_as_float(self, quote_atoms):
        return quote_atoms / self.quote_atoms_per_quote_unit

    def base_lots_and_price_to_quote_atoms(self, base_lots, price_in_ticks):
        return (
            base_lots
            * price_in_ticks
            * self.tick_size_in_quote_atoms_per_base_unit
            // self.num_base_lots_per_base_unit
        )

    def float_price_to_ticks_rounded_down(self, price):
        result = (
            price * self.raw_base_units_per_base_unit * self.quote_atoms_per_quote_unit
        ) / self.tick_size_in_quote_atoms_per_base_unit
        return int(result)

    def float_price_to_ticks_rounded_up(self, price):
        result = (
            price * self.raw_base_units_per_base_unit * self.quote_atoms_per_quote_unit
        ) / self.tick_size_in_quote_atoms_per_base_unit
        return int(result + 0.5)  # using 0.5 to achieve rounding up

    def ticks_to_float_price(self, ticks):
        return (
            ticks
            * self.tick_size_in_quote_atoms_per_base_unit
            / (self.quote_atoms_per_quote_unit * self.raw_base_units_per_base_unit)
        )

    def raw_base_units_per_base_lot(self):
        return self.base_atoms_per_base_lot / self.base_atoms_per_raw_base_unit

    def quote_units_per_raw_base_unit_per_tick(self):
        return self.tick_size_in_quote_atoms_per_base_unit / (
            self.quote_atoms_per_quote_unit * self.raw_base_units_per_base_unit
        )

    def order_to_quote_lots(self, base_lots, price_in_ticks):
        return (
            base_lots
            * price_in_ticks
            * self.tick_size_in_quote_atoms_per_base_unit
            // (self.num_base_lots_per_base_unit * self.quote_atoms_per_quote_lot)
        )

    def create_place_limit_order_instruction(
        self, limit_order_packet: Union[Limit, PostOnly], trader_pubkey: Pubkey
    ) -> Instruction:
        # TODO: Refactor these into a separate helper function
        log_account = Pubkey.find_program_address([b"log"], PROGRAM_ID)[0]
        seat = Pubkey.find_program_address(
            [
                b"seat",
                bytes(self.address),
                bytes(trader_pubkey),
            ],
            PROGRAM_ID,
        )[0]
        base_account = get_associated_token_address(trader_pubkey, self.base_mint)
        quote_account = get_associated_token_address(trader_pubkey, self.quote_mint)
        accounts = PlaceLimitOrderAccounts(
            phoenix_program=PROGRAM_ID,
            log_authority=log_account,
            market=self.address,
            trader=trader_pubkey,
            seat=seat,
            base_account=base_account,
            quote_account=quote_account,
            base_vault=self.base_vault,
            quote_vault=self.quote_vault,
        )
        place_limit_order_args = PlaceLimitOrderArgs(order_packet=limit_order_packet)

        return place_limit_order(
            place_limit_order_args,
            accounts,
        )

    def create_swap_instruction(
        self, ioc_order_packet: ImmediateOrCancel, trader_pubkey: Pubkey
    ) -> Instruction:
        log_account = Pubkey.find_program_address([b"log"], PROGRAM_ID)[0]
        base_account = get_associated_token_address(trader_pubkey, self.base_mint)
        quote_account = get_associated_token_address(trader_pubkey, self.quote_mint)
        accounts = SwapAccounts(
            phoenix_program=PROGRAM_ID,
            log_authority=log_account,
            market=self.address,
            trader=trader_pubkey,
            base_account=base_account,
            quote_account=quote_account,
            base_vault=self.base_vault,
            quote_vault=self.quote_vault,
        )
        swap_args = SwapArgs(order_packet=ioc_order_packet)
        return swap(
            swap_args,
            accounts,
        )



from __future__ import annotations
from . import (
    side,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class CancelOrderParamsJSON(typing.TypedDict):
    side: side.SideJSON
    price_in_ticks: int
    order_sequence_number: int


@dataclass
class CancelOrderParams:
    layout: typing.ClassVar = borsh.CStruct(
        "side" / side.layout,
        "price_in_ticks" / borsh.U64,
        "order_sequence_number" / borsh.U64,
    )
    side: side.SideKind
    price_in_ticks: int
    order_sequence_number: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "CancelOrderParams":
        return cls(
            side=side.from_decoded(obj.side),
            price_in_ticks=obj.price_in_ticks,
            order_sequence_number=obj.order_sequence_number,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "side": self.side.to_encodable(),
            "price_in_ticks": self.price_in_ticks,
            "order_sequence_number": self.order_sequence_number,
        }

    def to_json(self) -> CancelOrderParamsJSON:
        return {
            "side": self.side.to_json(),
            "price_in_ticks": self.price_in_ticks,
            "order_sequence_number": self.order_sequence_number,
        }

    @classmethod
    def from_json(cls, obj: CancelOrderParamsJSON) -> "CancelOrderParams":
        return cls(
            side=side.from_json(obj["side"]),
            price_in_ticks=obj["price_in_ticks"],
            order_sequence_number=obj["order_sequence_number"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class AuditLogHeaderJSON(typing.TypedDict):
    instruction: int
    sequence_number: int
    timestamp: int
    slot: int
    market: str
    signer: str
    total_events: int


@dataclass
class AuditLogHeader:
    layout: typing.ClassVar = borsh.CStruct(
        "instruction" / borsh.U8,
        "sequence_number" / borsh.U64,
        "timestamp" / borsh.I64,
        "slot" / borsh.U64,
        "market" / BorshPubkey,
        "signer" / BorshPubkey,
        "total_events" / borsh.U16,
    )
    instruction: int
    sequence_number: int
    timestamp: int
    slot: int
    market: Pubkey
    signer: Pubkey
    total_events: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "AuditLogHeader":
        return cls(
            instruction=obj.instruction,
            sequence_number=obj.sequence_number,
            timestamp=obj.timestamp,
            slot=obj.slot,
            market=obj.market,
            signer=obj.signer,
            total_events=obj.total_events,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "instruction": self.instruction,
            "sequence_number": self.sequence_number,
            "timestamp": self.timestamp,
            "slot": self.slot,
            "market": self.market,
            "signer": self.signer,
            "total_events": self.total_events,
        }

    def to_json(self) -> AuditLogHeaderJSON:
        return {
            "instruction": self.instruction,
            "sequence_number": self.sequence_number,
            "timestamp": self.timestamp,
            "slot": self.slot,
            "market": str(self.market),
            "signer": str(self.signer),
            "total_events": self.total_events,
        }

    @classmethod
    def from_json(cls, obj: AuditLogHeaderJSON) -> "AuditLogHeader":
        return cls(
            instruction=obj["instruction"],
            sequence_number=obj["sequence_number"],
            timestamp=obj["timestamp"],
            slot=obj["slot"],
            market=Pubkey.from_string(obj["market"]),
            signer=Pubkey.from_string(obj["signer"]),
            total_events=obj["total_events"],
        )



import base58
import struct
from typing import List, Optional
from phoenix.program_id import PROGRAM_ID
from phoenix.types.audit_log_header import AuditLogHeader
from phoenix.types.fill_summary_event import FillSummaryEvent
from phoenix.types.phoenix_market_event import PhoenixMarketEventKind, from_decoded
from phoenix.types.phoenix_market_event import layout as phoenix_market_event_layout
from phoenix.types.place_event import PlaceEvent
from phoenix.types.reduce_event import ReduceEvent
from phoenix.types.evict_event import EvictEvent
from phoenix.types.expired_order_event import ExpiredOrderEvent
from phoenix.types.fee_event import FeeEvent
from phoenix.types.time_in_force_event import TimeInForceEvent
from phoenix.types.fill_event import FillEvent
from solders.transaction_status import EncodedConfirmedTransactionWithStatusMeta
from solders.pubkey import Pubkey

LOG_INSTRUCTION_DISCRIMINATOR = 15


class PhoenixEventsFromInstruction:
    def __init__(self, header: AuditLogHeader, events: List[PhoenixMarketEventKind]):
        self.header = header
        self.events = events


class PhoenixTransaction:
    def __init__(
        self,
        events_from_instructions: List[PhoenixEventsFromInstruction],
        signature: Optional[str] = None,
        txReceived: bool = False,
        txFailed: bool = False,
    ):
        self.events_from_instructions = events_from_instructions
        self.signature = signature
        self.txReceived = txReceived
        self.txFailed = txFailed


class PhoenixEvents:
    def __init__(self, events: List[PhoenixMarketEventKind]):
        self.events = events


def get_phoenix_events_from_confirmed_transaction_with_meta(
    tx_data: EncodedConfirmedTransactionWithStatusMeta,
) -> PhoenixTransaction:
    meta = tx_data.transaction.meta

    if not meta:
        return PhoenixTransaction([], txReceived=False, txFailed=True)

    if meta.err is not None:
        return PhoenixTransaction([], txReceived=True, txFailed=True)

    inner_ixs = meta.inner_instructions
    if not inner_ixs or not tx_data.transaction or not tx_data.slot:
        return PhoenixTransaction([], txReceived=True, txFailed=True)

    data_array: List[bytes] = []
    account_keys = tx_data.transaction.transaction.message.account_keys
    if meta.loaded_addresses is not None:
        account_keys += meta.loaded_addresses.writable
        account_keys += meta.loaded_addresses.readonly

    for ix in inner_ixs:
        for inner in ix.instructions:
            if account_keys[inner.program_id_index] != PROGRAM_ID:
                continue
            raw_data = base58.b58decode(inner.data)
            if raw_data[0] == LOG_INSTRUCTION_DISCRIMINATOR:
                data_array.append(raw_data[1:])

    instructions = [get_phoenix_events_from_log_data(data) for data in data_array]

    return PhoenixTransaction(
        instructions,
        signature=tx_data.transaction.transaction.signatures[0],
        txReceived=True,
        txFailed=False,
    )


def get_phoenix_events_from_log_data(data: bytes) -> PhoenixEventsFromInstruction:
    reader_index = 0

    # Read the first byte
    # A byte of 1 identifies a header event
    (byte,) = struct.unpack_from("B", data, reader_index)
    reader_index += 1
    if byte != 1:
        raise Exception("early Unexpected event")

    header = AuditLogHeader.layout.parse(data[reader_index:])
    # Create the length buffer
    length_buffer = struct.pack("I", header.total_events)
    # Coerce the buffer for Borsh-encoded vector decoding
    events_data = length_buffer + data[AuditLogHeader.layout.sizeof() + 1 :]

    events = decode_phoenix_events(events_data)

    return PhoenixEventsFromInstruction(header, events)


EVENT_CLASSES = {
    "Uninitialized": None,  # No event class for "Uninitialized"
    "Header": AuditLogHeader,
    "Fill": FillEvent,
    "Place": PlaceEvent,
    "Reduce": ReduceEvent,
    "Evict": EvictEvent,
    "FillSummary": FillSummaryEvent,
    "Fee": FeeEvent,
    "TimeInForce": TimeInForceEvent,
    "ExpiredOrder": ExpiredOrderEvent,
}


def decode_phoenix_events(data: bytes) -> List[PhoenixMarketEventKind]:
    # Decode the length of the events array
    num_of_events = int.from_bytes(data[:4], "little")
    offset = 4
    events = []

    while offset < len(data):
        # Deserializing each event based on the layout
        event = phoenix_market_event_layout.parse(data[offset:])
        # Get the corresponding class, parse and adjust the offset
        event_type = list(event.keys())[0]
        event_class = EVENT_CLASSES[event_type]
        event_instance = from_decoded(event)
        events.append(event_instance)
        event_size = event_class.layout.sizeof() + 1
        offset += event_size
    assert (
        len(events) == num_of_events
    ), "Decoding events from transaction: Mismatch in event length"

    return events


def read_public_key(data: bytes) -> Pubkey:
    # Extract the 32 bytes for the public key
    pubkey_bytes = data[:32]

    # Convert the bytes to a Pubkey
    pubkey = Pubkey.from_bytes(pubkey_bytes)

    return pubkey



from __future__ import annotations
from . import (
    audit_log_header,
    fill_event,
    reduce_event,
    place_event,
    evict_event,
    fill_summary_event,
    fee_event,
    time_in_force_event,
    expired_order_event,
)
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh

HeaderJSONValue = tuple[audit_log_header.AuditLogHeaderJSON]
FillJSONValue = tuple[fill_event.FillEventJSON]
PlaceJSONValue = tuple[place_event.PlaceEventJSON]
ReduceJSONValue = tuple[reduce_event.ReduceEventJSON]
EvictJSONValue = tuple[evict_event.EvictEventJSON]
FillSummaryJSONValue = tuple[fill_summary_event.FillSummaryEventJSON]
FeeJSONValue = tuple[fee_event.FeeEventJSON]
TimeInForceJSONValue = tuple[time_in_force_event.TimeInForceEventJSON]
ExpiredOrderJSONValue = tuple[expired_order_event.ExpiredOrderEventJSON]
HeaderValue = tuple[audit_log_header.AuditLogHeader]
FillValue = tuple[fill_event.FillEvent]
PlaceValue = tuple[place_event.PlaceEvent]
ReduceValue = tuple[reduce_event.ReduceEvent]
EvictValue = tuple[evict_event.EvictEvent]
FillSummaryValue = tuple[fill_summary_event.FillSummaryEvent]
FeeValue = tuple[fee_event.FeeEvent]
TimeInForceValue = tuple[time_in_force_event.TimeInForceEvent]
ExpiredOrderValue = tuple[expired_order_event.ExpiredOrderEvent]


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class HeaderJSON(typing.TypedDict):
    value: HeaderJSONValue
    kind: typing.Literal["Header"]


class FillJSON(typing.TypedDict):
    value: FillJSONValue
    kind: typing.Literal["Fill"]


class PlaceJSON(typing.TypedDict):
    value: PlaceJSONValue
    kind: typing.Literal["Place"]


class ReduceJSON(typing.TypedDict):
    value: ReduceJSONValue
    kind: typing.Literal["Reduce"]


class EvictJSON(typing.TypedDict):
    value: EvictJSONValue
    kind: typing.Literal["Evict"]


class FillSummaryJSON(typing.TypedDict):
    value: FillSummaryJSONValue
    kind: typing.Literal["FillSummary"]


class FeeJSON(typing.TypedDict):
    value: FeeJSONValue
    kind: typing.Literal["Fee"]


class TimeInForceJSON(typing.TypedDict):
    value: TimeInForceJSONValue
    kind: typing.Literal["TimeInForce"]


class ExpiredOrderJSON(typing.TypedDict):
    value: ExpiredOrderJSONValue
    kind: typing.Literal["ExpiredOrder"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class Header:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Header"
    value: HeaderValue

    def to_json(self) -> HeaderJSON:
        return HeaderJSON(
            kind="Header",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Header": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Fill:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Fill"
    value: FillValue

    def to_json(self) -> FillJSON:
        return FillJSON(
            kind="Fill",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Fill": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Place:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Place"
    value: PlaceValue

    def to_json(self) -> PlaceJSON:
        return PlaceJSON(
            kind="Place",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Place": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Reduce:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "Reduce"
    value: ReduceValue

    def to_json(self) -> ReduceJSON:
        return ReduceJSON(
            kind="Reduce",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Reduce": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Evict:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "Evict"
    value: EvictValue

    def to_json(self) -> EvictJSON:
        return EvictJSON(
            kind="Evict",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Evict": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class FillSummary:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "FillSummary"
    value: FillSummaryValue

    def to_json(self) -> FillSummaryJSON:
        return FillSummaryJSON(
            kind="FillSummary",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "FillSummary": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class Fee:
    discriminator: typing.ClassVar = 7
    kind: typing.ClassVar = "Fee"
    value: FeeValue

    def to_json(self) -> FeeJSON:
        return FeeJSON(
            kind="Fee",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "Fee": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class TimeInForce:
    discriminator: typing.ClassVar = 8
    kind: typing.ClassVar = "TimeInForce"
    value: TimeInForceValue

    def to_json(self) -> TimeInForceJSON:
        return TimeInForceJSON(
            kind="TimeInForce",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "TimeInForce": {
                "item_0": self.value[0].to_encodable(),
            },
        }


@dataclass
class ExpiredOrder:
    discriminator: typing.ClassVar = 9
    kind: typing.ClassVar = "ExpiredOrder"
    value: ExpiredOrderValue

    def to_json(self) -> ExpiredOrderJSON:
        return ExpiredOrderJSON(
            kind="ExpiredOrder",
            value=(self.value[0].to_json(),),
        )

    def to_encodable(self) -> dict:
        return {
            "ExpiredOrder": {
                "item_0": self.value[0].to_encodable(),
            },
        }


PhoenixMarketEventKind = typing.Union[
    Uninitialized,
    Header,
    Fill,
    Place,
    Reduce,
    Evict,
    FillSummary,
    Fee,
    TimeInForce,
    ExpiredOrder,
]
PhoenixMarketEventJSON = typing.Union[
    UninitializedJSON,
    HeaderJSON,
    FillJSON,
    PlaceJSON,
    ReduceJSON,
    EvictJSON,
    FillSummaryJSON,
    FeeJSON,
    TimeInForceJSON,
    ExpiredOrderJSON,
]


def from_decoded(obj: dict) -> PhoenixMarketEventKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "Header" in obj:
        val = obj["Header"]
        return Header((audit_log_header.AuditLogHeader.from_decoded(val["item_0"]),))
    if "Fill" in obj:
        val = obj["Fill"]
        return Fill((fill_event.FillEvent.from_decoded(val["item_0"]),))
    if "Place" in obj:
        val = obj["Place"]
        return Place((place_event.PlaceEvent.from_decoded(val["item_0"]),))
    if "Reduce" in obj:
        val = obj["Reduce"]
        return Reduce((reduce_event.ReduceEvent.from_decoded(val["item_0"]),))
    if "Evict" in obj:
        val = obj["Evict"]
        return Evict((evict_event.EvictEvent.from_decoded(val["item_0"]),))
    if "FillSummary" in obj:
        val = obj["FillSummary"]
        return FillSummary(
            (fill_summary_event.FillSummaryEvent.from_decoded(val["item_0"]),)
        )
    if "Fee" in obj:
        val = obj["Fee"]
        return Fee((fee_event.FeeEvent.from_decoded(val["item_0"]),))
    if "TimeInForce" in obj:
        val = obj["TimeInForce"]
        return TimeInForce(
            (time_in_force_event.TimeInForceEvent.from_decoded(val["item_0"]),)
        )
    if "ExpiredOrder" in obj:
        val = obj["ExpiredOrder"]
        return ExpiredOrder(
            (expired_order_event.ExpiredOrderEvent.from_decoded(val["item_0"]),)
        )
    raise ValueError("Invalid enum object")


def from_json(obj: PhoenixMarketEventJSON) -> PhoenixMarketEventKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "Header":
        header_json_value = typing.cast(HeaderJSONValue, obj["value"])
        return Header(
            (audit_log_header.AuditLogHeader.from_json(header_json_value[0]),)
        )
    if obj["kind"] == "Fill":
        fill_json_value = typing.cast(FillJSONValue, obj["value"])
        return Fill((fill_event.FillEvent.from_json(fill_json_value[0]),))
    if obj["kind"] == "Place":
        place_json_value = typing.cast(PlaceJSONValue, obj["value"])
        return Place((place_event.PlaceEvent.from_json(place_json_value[0]),))
    if obj["kind"] == "Reduce":
        reduce_json_value = typing.cast(ReduceJSONValue, obj["value"])
        return Reduce((reduce_event.ReduceEvent.from_json(reduce_json_value[0]),))
    if obj["kind"] == "Evict":
        evict_json_value = typing.cast(EvictJSONValue, obj["value"])
        return Evict((evict_event.EvictEvent.from_json(evict_json_value[0]),))
    if obj["kind"] == "FillSummary":
        fill_summary_json_value = typing.cast(FillSummaryJSONValue, obj["value"])
        return FillSummary(
            (fill_summary_event.FillSummaryEvent.from_json(fill_summary_json_value[0]),)
        )
    if obj["kind"] == "Fee":
        fee_json_value = typing.cast(FeeJSONValue, obj["value"])
        return Fee((fee_event.FeeEvent.from_json(fee_json_value[0]),))
    if obj["kind"] == "TimeInForce":
        time_in_force_json_value = typing.cast(TimeInForceJSONValue, obj["value"])
        return TimeInForce(
            (
                time_in_force_event.TimeInForceEvent.from_json(
                    time_in_force_json_value[0]
                ),
            )
        )
    if obj["kind"] == "ExpiredOrder":
        expired_order_json_value = typing.cast(ExpiredOrderJSONValue, obj["value"])
        return ExpiredOrder(
            (
                expired_order_event.ExpiredOrderEvent.from_json(
                    expired_order_json_value[0]
                ),
            )
        )
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "Header" / borsh.CStruct("item_0" / audit_log_header.AuditLogHeader.layout),
    "Fill" / borsh.CStruct("item_0" / fill_event.FillEvent.layout),
    "Place" / borsh.CStruct("item_0" / place_event.PlaceEvent.layout),
    "Reduce" / borsh.CStruct("item_0" / reduce_event.ReduceEvent.layout),
    "Evict" / borsh.CStruct("item_0" / evict_event.EvictEvent.layout),
    "FillSummary"
    / borsh.CStruct("item_0" / fill_summary_event.FillSummaryEvent.layout),
    "Fee" / borsh.CStruct("item_0" / fee_event.FeeEvent.layout),
    "TimeInForce"
    / borsh.CStruct("item_0" / time_in_force_event.TimeInForceEvent.layout),
    "ExpiredOrder"
    / borsh.CStruct("item_0" / expired_order_event.ExpiredOrderEvent.layout),
)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class WithdrawFundsArgs(typing.TypedDict):
    withdraw_funds_params: types.withdraw_params.WithdrawParams


layout = borsh.CStruct(
    "withdraw_funds_params" / types.withdraw_params.WithdrawParams.layout
)


class WithdrawFundsAccounts(typing.TypedDict):
    phoenix_program: Pubkey
    log_authority: Pubkey
    market: Pubkey
    trader: Pubkey
    base_account: Pubkey
    quote_account: Pubkey
    base_vault: Pubkey
    quote_vault: Pubkey


def withdraw_funds(
    args: WithdrawFundsArgs,
    accounts: WithdrawFundsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["phoenix_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["log_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["trader"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["base_account"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["quote_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x0c"
    encoded_args = layout.build(
        {
            "withdraw_funds_params": args["withdraw_funds_params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import asyncio
import traceback
from dataclasses import dataclass
import base58
import json
from typing import Any, Callable, Dict, Iterator, Tuple, TypedDict, Union, List
from uuid import uuid4
import time
import traceback
from asyncstdlib import enumerate
import websockets
from phoenix.instructions import cancel_all_orders, withdraw_funds
from phoenix.instructions import cancel_multiple_orders_by_id
from phoenix.instructions import cancel_all_orders_with_free_funds
from phoenix.instructions import cancel_multiple_orders_by_id_with_free_funds
from phoenix.instructions.cancel_all_orders import CancelAllOrdersAccounts
from phoenix.instructions.cancel_all_orders_with_free_funds import (
    CancelAllOrdersWithFreeFundsAccounts,
)
from phoenix.instructions.cancel_multiple_orders_by_id import (
    CancelMultipleOrdersByIdAccounts,
    CancelMultipleOrdersByIdArgs,
)
from phoenix.instructions.withdraw_funds import (
    WithdrawFundsAccounts,
    WithdrawFundsArgs,
    withdraw_funds,
)
from phoenix.instructions.cancel_multiple_orders_by_id_with_free_funds import (
    CancelMultipleOrdersByIdWithFreeFundsAccounts,
    CancelMultipleOrdersByIdWithFreeFundsArgs,
)
from phoenix.market_metadata import MarketMetadata
from phoenix.order_subscribe_response import (
    CancelledOrder,
    FilledOrder,
    OpenOrder,
    OrderSubscribeError,
    OrderSubscribeResponse,
)
from phoenix.program_id import PROGRAM_ID
from phoenix.types import self_trade_behavior
from phoenix.types.cancel_multiple_orders_by_id_params import (
    CancelMultipleOrdersByIdParams,
)
from phoenix.types.cancel_order_params import CancelOrderParams
from phoenix.types.order_packet import (
    ImmediateOrCancel,
    ImmediateOrCancelValue,
    Limit,
    LimitValue,
    OrderPacketKind,
    PostOnly,
    PostOnlyValue,
)
from spl.token.instructions import get_associated_token_address
from phoenix.types.side import Ask, Bid, SideKind, from_order_sequence_number
from phoenix.events import (
    PhoenixTransaction,
    get_phoenix_events_from_confirmed_transaction_with_meta,
    get_phoenix_events_from_log_data,
)
from solana.rpc.commitment import Commitment
from solana.rpc.async_api import AsyncClient
from solana.rpc.types import TxOpts
from solders.pubkey import Pubkey
from solders.hash import Hash
from solders.signature import Signature
from solders.rpc.config import RpcTransactionLogsFilterMentions
from solana.transaction import Instruction, Transaction, Keypair
from solana.rpc.websocket_api import connect
from phoenix.types.fifo_order_id import FIFOOrderId
from phoenix.types.withdraw_params import WithdrawParams
from jsonrpcclient import request

from .market import DEFAULT_L2_LADDER_DEPTH, ActiveOrder, Ladder, Market


class ExecutableOrder:
    def __init__(
        self,
        order_packet: OrderPacketKind,
        market_pubkey: Pubkey,
    ):
        self.order_packet = order_packet
        self.market_pubkey = market_pubkey


class PhoenixOrder:
    def __init__(
        self,
        exchange_order_id: FIFOOrderId,
        base_lots_remaining: int,
    ):
        self.order_id = exchange_order_id.to_int()
        self.base_lots_remaining = base_lots_remaining

    def __repr__(self):
        return f"PhoenixOrder({self.order_id}, {self.base_lots_remaining})"


class PhoenixResponse:
    def __init__(
        self,
        signature: Signature,
        sequence_number: int,
        client_orders_map: Dict[int, PhoenixOrder],
        cancelled_orders: List[PhoenixOrder] = None,
    ):
        self.signature = signature
        self.sequence_number = sequence_number
        self.client_orders_map = client_orders_map
        if cancelled_orders is None:
            self.cancelled_orders = []
        else:
            self.cancelled_orders = cancelled_orders


class PhoenixClient:
    def __init__(
        self,
        endpoint="api.mainnet-beta.solana.com",
        custom_url=None,
        custom_ws_url=None,
        commitment: Commitment = "confirmed",
        encoding: str = "base64+zstd",
    ):
        if custom_url is None:
            self.endpoint = f"https://{endpoint.split('://')[-1]}"
        else:
            self.endpoint = custom_url
        self.client = AsyncClient(self.endpoint)

        if custom_ws_url is None:
            self.ws_endpoint = f"wss://{endpoint.split('://')[-1]}"
        else:
            self.ws_endpoint = custom_ws_url

        self.commitment = commitment
        self.encoding = encoding

        self.markets = {}

        self.__slot = -1
        self.__subscribed_to_slot = False

    async def add_market(self, market_pubkey: Pubkey):
        market_bytes = await self.client.get_account_info(
            market_pubkey, self.commitment, self.encoding
        )
        self.markets[market_pubkey] = Market.deserialize_market_data(
            market_pubkey, market_bytes.value.data
        ).metadata

    async def get_l2_book(self, market_pubkey: Pubkey, levels=DEFAULT_L2_LADDER_DEPTH):
        market_account = await self.client.get_account_info(
            market_pubkey, self.commitment, self.encoding
        )
        slot = market_account.context.slot
        market = Market.deserialize_market_data(
            market_pubkey, market_account.value.data
        )
        unix_timestamp = (await self.client.get_block_time(slot)).value
        return market.get_ui_ladder(slot, unix_timestamp, levels)

    """
    Returns a list of PhoenixOrders that are active for the given market and trader

    Params:
    
    market_pubkey: Pubkey of the market to get orders for
    trader_pubkey: Pubkey of the trader to get orders for
    side: SideKind (Bid or Ask) of the orders to get. You may also pass in the strings "buy" | "bid" or "sell" | "ask" (optional)
    ids: List of order ids (as int) to get. If None, all orders will be returned (optional)
    market: Market object. If None, the market will be fetched from the client (optional)
    """

    async def get_active_orders(
        self,
        market_pubkey: Pubkey,
        trader_pubkey: Pubkey,
        side: SideKind | str | None = None,
        ids: List[int] | None = None,
        market: Market | None = None,
    ) -> List[ActiveOrder]:
        if side is not None:
            if isinstance(side, str):
                side = side.lower().replace("bid", "buy").replace("ask", "sell")
                if side not in ["buy", "sell"]:
                    raise Exception(f"Invalid side: {side}")
            elif side == Bid():
                side = "buy"
            elif side == Ask():
                side == "sell"
            else:
                raise Exception(f"Invalid side: {side}")
        if market is None:
            market_account = await self.client.get_account_info(
                market_pubkey, self.commitment, self.encoding
            )
            market = Market.deserialize_market_data(
                market_pubkey, market_account.value.data
            )
        active_orders = market.get_active_orders(trader=trader_pubkey)
        if side is not None:
            active_orders = list(filter(lambda o: o.side == side, active_orders))
        if ids is not None:
            ids = set(ids)
            active_orders = list(filter(lambda o: o.order_id in ids, active_orders))
        return active_orders

    async def order_subscribe_with_default_client(
        self, market_pubkey: Pubkey, trader_pubkey: Pubkey
    ) -> Iterator[List[OrderSubscribeResponse]]:
        error_count = 0
        reconnection_count = 0
        if market_pubkey not in self.markets:
            await self.add_market(market_pubkey)
        while True:
            reconnection_count += 1
            async with connect(self.ws_endpoint) as websocket:
                await websocket.logs_subscribe(
                    commitment=self.commitment,
                    filter_=RpcTransactionLogsFilterMentions(PROGRAM_ID),
                )
                await websocket.recv()
                async for msg in websocket:
                    if msg[0].result.value.err != None:
                        continue
                    try:
                        sig = msg[0].result.value.signature
                        transaction = None
                        attempts = 0
                        while transaction is None and attempts < 5:
                            attempts += 1
                            tx_response = await self.client.get_transaction(
                                sig,
                                "json",
                                self.commitment,
                                max_supported_transaction_version=0,
                            )
                            transaction = tx_response.value
                        if transaction is None:
                            error_count += 1
                            yield (
                                [
                                    OrderSubscribeError(
                                        message=f"Transaction not found: {sig}",
                                        error_count=error_count,
                                        reconnection_count=reconnection_count,
                                        reconnected=False,
                                        traceback=None,
                                        exception=None,
                                    )
                                ]
                            )
                            continue
                        phoenix_transaction = (
                            get_phoenix_events_from_confirmed_transaction_with_meta(
                                transaction
                            )
                        )
                        response = self.__get_response_from_phoenix_transaction(
                            phoenix_transaction, market_pubkey, trader_pubkey
                        )
                        if len(response) > 0:
                            yield response
                    except websockets.exceptions.ConnectionClosed as e:
                        print("Connection closed unexpectedly. Reconnecting...")
                        err = OrderSubscribeError(
                            message="Websocked connection closed",
                            error_count=error_count,
                            reconnection_count=reconnection_count,
                            reconnected=True,
                            traceback=traceback.extract_stack(),
                            exception=e,
                        )
                        yield [err]
                        error_count += 1
                        break
                    except Exception as e:
                        err = OrderSubscribeError(
                            message=f"Error processing transaction event: {e}",
                            error_count=error_count,
                            reconnection_count=reconnection_count,
                            reconnected=False,
                            traceback=traceback.extract_stack(),
                            exception=e,
                        )
                        yield [err]
                        error_count += 1

    async def order_subscribe(
        self,
        market_pubkey: Pubkey,
        trader_pubkey: Pubkey,
        commitment: Commitment | None = None,
    ) -> Iterator[List[OrderSubscribeResponse]]:
        error_count = 0
        reconnection_count = 0
        if market_pubkey not in self.markets:
            await self.add_market(market_pubkey)
        while True:
            reconnection_count += 1
            async with websockets.connect(self.ws_endpoint + "/whirligig") as websocket:
                commitment = commitment if commitment != None else self.commitment
                try:
                    transaction_subscribe = request(
                        "transactionSubscribe",
                        params=[
                            {
                                "mentions": [str(market_pubkey)],
                                "failed": False,
                                "vote": False,
                            },
                            {
                                "commitment": str(commitment),
                            },
                        ],
                    )
                    await websocket.send(json.dumps(transaction_subscribe))
                    # Ignore the first message
                    await websocket.recv()
                    while True:
                        try:
                            response = self.__process_transaction_event(
                                json.loads(await websocket.recv()),
                                market_pubkey,
                                trader_pubkey,
                            )
                            if len(response) > 0:
                                yield response
                        except websockets.exceptions.ConnectionClosed as e:
                            print("Connection closed unexpectedly. Reconnecting...")
                            err = OrderSubscribeError(
                                message="Websocked connection closed",
                                error_count=error_count,
                                reconnection_count=reconnection_count,
                                reconnected=True,
                                traceback=traceback.extract_stack(),
                                exception=e,
                            )
                            yield [err]
                            error_count += 1
                            break
                        except Exception as e:
                            err = OrderSubscribeError(
                                message=f"Error processing transaction event: {e}",
                                error_count=error_count,
                                reconnection_count=reconnection_count,
                                reconnected=False,
                                traceback=traceback.extract_stack(),
                                exception=e,
                            )
                            yield [err]
                            error_count += 1

                except Exception as e:
                    err = OrderSubscribeError(
                        message=f"Error processing transaction event: {e}",
                        error_count=error_count,
                        reconnection_count=reconnection_count,
                        reconnected=False,
                        traceback=traceback.extract_stack(),
                        exception=e,
                    )
                    yield [err]
                    error_count += 1

    def __process_transaction_event(
        self, response, market_pubkey: Pubkey, trader_pubkey: Pubkey
    ) -> [OrderSubscribeResponse]:
        payload = response["params"]["result"]["value"]
        meta = payload["meta"]
        loaded_addresses = meta.get("loadedAddresses", {"readonly": [], "writable": []})
        tx_message = payload["transaction"]["message"]
        message = tx_message[-1] if isinstance(tx_message, list) else tx_message
        base_account_keys = list(
            map(
                lambda l: Pubkey.from_bytes(bytes(l)),
                message["accountKeys"][1:],
            )
        )
        lookup_table_keys = list(
            map(
                Pubkey.from_string,
                loaded_addresses["writable"] + loaded_addresses["readonly"],
            )
        )
        account_keys = base_account_keys + lookup_table_keys
        data_array: List[bytes] = []
        for inner_ix in meta["innerInstructions"]:
            for ix in inner_ix["instructions"]:
                program = account_keys[ix["programIdIndex"]]
                if program != PROGRAM_ID:
                    continue
                data = base58.b58decode(ix["data"])
                if data[0] == 15:
                    data_array.append(data[1:])

        instructions = [get_phoenix_events_from_log_data(data) for data in data_array]

        phoenix_tx = PhoenixTransaction(
            instructions,
            signature=Signature.from_bytes(payload["transaction"]["signatures"][1]),
            txReceived=True,
            txFailed=False,
        )
        return self.__get_response_from_phoenix_transaction(
            phoenix_tx, market_pubkey=market_pubkey, trader_pubkey=trader_pubkey
        )

    def __get_response_from_phoenix_transaction(
        self, phoenix_tx, market_pubkey, trader_pubkey
    ):
        meta: MarketMetadata = self.markets[market_pubkey]
        response = []
        for event in phoenix_tx.events_from_instructions:
            header = event.header
            if header.market != market_pubkey:
                continue
            for e in event.events:
                if e.kind == "Fill":
                    fill = e.value[0]
                    if fill.maker_id == trader_pubkey or header.signer == trader_pubkey:
                        fifo_order_id = FIFOOrderId(
                            order_sequence_number=fill.order_sequence_number,
                            price_in_ticks=fill.price_in_ticks,
                        )
                        price = meta.ticks_to_float_price(fill.price_in_ticks)
                        quantity_filled = meta.base_lots_to_raw_base_units_as_float(
                            fill.base_lots_filled
                        )
                        quantity_remaining = meta.base_lots_to_raw_base_units_as_float(
                            fill.base_lots_remaining
                        )
                        filled_order = FilledOrder(
                            order_id=fifo_order_id.to_int(),
                            market_pubkey=market_pubkey,
                            slot=header.slot,
                            unix_timestamp_in_seconds=header.timestamp,
                            sequence_number=header.sequence_number,
                            taker_side=from_order_sequence_number(
                                fill.order_sequence_number
                            ).opposite(),
                            price=price,
                            quantity_filled=quantity_filled,
                            quantity_remaining=quantity_remaining,
                            taker_pubkey=header.signer,
                            maker_pubkey=fill.maker_id,
                        )
                        response.append(filled_order)
                elif e.kind == "Place":
                    place = e.value[0]
                    if header.signer == trader_pubkey:
                        fifo_order_id = FIFOOrderId(
                            order_sequence_number=place.order_sequence_number,
                            price_in_ticks=place.price_in_ticks,
                        )
                        price = meta.ticks_to_float_price(place.price_in_ticks)
                        quantity_placed = meta.base_lots_to_raw_base_units_as_float(
                            place.base_lots_placed
                        )
                        open_order = OpenOrder(
                            order_id=fifo_order_id.to_int(),
                            client_order_id=place.client_order_id,
                            market_pubkey=market_pubkey,
                            slot=header.slot,
                            unix_timestamp_in_seconds=header.timestamp,
                            sequence_number=header.sequence_number,
                            side=from_order_sequence_number(
                                place.order_sequence_number
                            ),
                            price=price,
                            quantity_placed=quantity_placed,
                            maker_pubkey=trader_pubkey,
                        )
                        response.append(open_order)
                elif e.kind == "Reduce":
                    reduce = e.value[0]
                    if header.signer == trader_pubkey:
                        fifo_order_id = FIFOOrderId(
                            order_sequence_number=reduce.order_sequence_number,
                            price_in_ticks=reduce.price_in_ticks,
                        )
                        price = meta.ticks_to_float_price(reduce.price_in_ticks)
                        quantity_remaining = meta.base_lots_to_raw_base_units_as_float(
                            reduce.base_lots_remaining
                        )
                        quantity_removed = meta.base_lots_to_raw_base_units_as_float(
                            reduce.base_lots_removed
                        )
                        cancelled_order = CancelledOrder(
                            order_id=fifo_order_id.to_int(),
                            market_pubkey=market_pubkey,
                            slot=header.slot,
                            unix_timestamp_in_seconds=header.timestamp,
                            sequence_number=header.sequence_number,
                            side=from_order_sequence_number(
                                reduce.order_sequence_number
                            ),
                            price=price,
                            quantity_remaining=quantity_remaining,
                            quantity_removed=quantity_removed,
                            maker_pubkey=trader_pubkey,
                        )
                        response.append(cancelled_order)
                elif e.kind == "ExpiredOrder":
                    expired = e.value[0]
                    if header.signer == trader_pubkey:
                        fifo_order_id = FIFOOrderId(
                            order_sequence_number=expired.order_sequence_number,
                            price_in_ticks=expired.price_in_ticks,
                        )
                        price = meta.ticks_to_float_price(expired.price_in_ticks)
                        quantity_remaining = meta.base_lots_to_raw_base_units_as_float(
                            expired.base_lots_remaining
                        )
                        quantity_removed = meta.base_lots_to_raw_base_units_as_float(
                            expired.base_lots_removed
                        )
                        cancelled_order = CancelledOrder(
                            order_id=fifo_order_id.to_int(),
                            market_pubkey=market_pubkey,
                            slot=header.slot,
                            unix_timestamp_in_seconds=header.timestamp,
                            sequence_number=header.sequence_number,
                            side=from_order_sequence_number(
                                expired.order_sequence_number
                            ),
                            price=price,
                            quantity_remaining=quantity_remaining,
                            quantity_removed=quantity_removed,
                            maker_pubkey=trader_pubkey,
                        )
                        response.append(cancelled_order)
                elif e.kind == "Evict":
                    evicted = e.value[0]
                    if header.signer == trader_pubkey:
                        fifo_order_id = FIFOOrderId(
                            order_sequence_number=evicted.order_sequence_number,
                            price_in_ticks=evicted.price_in_ticks,
                        )
                        price = meta.ticks_to_float_price(evicted.price_in_ticks)
                        quantity_remaining = meta.base_lots_to_raw_base_units_as_float(
                            evicted.base_lots_remaining
                        )
                        quantity_removed = meta.base_lots_to_raw_base_units_as_float(
                            evicted.base_lots_removed
                        )
                        cancelled_order = CancelledOrder(
                            order_id=fifo_order_id.to_int(),
                            market_pubkey=market_pubkey,
                            slot=header.slot,
                            unix_timestamp_in_seconds=header.timestamp,
                            sequence_number=header.sequence_number,
                            side=from_order_sequence_number(
                                evicted.order_sequence_number
                            ),
                            price=price,
                            quantity_remaining=quantity_remaining,
                            quantity_removed=quantity_removed,
                            maker_pubkey=trader_pubkey,
                        )
                        response.append(cancelled_order)
        return response

    async def market_subscribe(
        self, market_pubkey: Pubkey, handle_market: Callable[[Market], Any]
    ):
        await asyncio.gather(
            self.__slot_subscribe(),
            self.__market_subscribe(market_pubkey, handle_market),
        )

    async def l2_orderbook_subscribe(
        self,
        market_pubkey: Pubkey,
        handle_l2_orderbook: Callable[[Ladder], Any],
        depth=DEFAULT_L2_LADDER_DEPTH,
    ):
        async def handle_market(market: Market):
            ladder = market.get_ui_ladder(
                slot=self.__slot, unix_timestamp=int(time.time()), levels=depth
            )
            await handle_l2_orderbook(ladder)

        await self.market_subscribe(market_pubkey, handle_market)

    async def __market_subscribe(
        self, market_pubkey: Pubkey, handle_market: Callable[[Market], Any]
    ):
        async with connect(self.ws_endpoint) as websocket:
            await websocket.account_subscribe(
                market_pubkey, self.commitment, self.encoding
            )
            first_resp = await websocket.recv()
            subscription_id = first_resp[0].result
            async for _, msg in enumerate(websocket):
                try:
                    market = Market.deserialize_market_data(
                        market_pubkey, msg[0].result.value.data
                    )
                    await handle_market(market)
                except Exception as e:
                    print(f"WARNING: Failed to parse message in market subscribe: {e}")
                    print(traceback.format_exc())

            await websocket.account_unsubscribe(subscription_id)

    async def __slot_subscribe(self):
        if self.__subscribed_to_slot:
            return
        self.__subscribed_to_slot = True
        print("Subscribing to slot")
        async with connect(self.ws_endpoint) as websocket:
            await websocket.slot_subscribe()
            first_resp = await websocket.recv()
            subscription_id = first_resp[0].result
            async for _, msg in enumerate(websocket):
                try:
                    self.slot = msg[0].result.slot
                except Exception as e:
                    print(f"WARNING: Failed to parse message: {e}")
            await websocket.slot_unsubscribe(subscription_id)
        self.__subscribed_to_slot = False

    async def close(self):
        await self.client.close()

    def get_limit_order_packet(
        self,
        market_pubkey: Pubkey,
        side: SideKind,
        price_in_quote_units: float,
        size_in_base_units: float,
        client_order_id: int = None,
        self_trade_behavior: self_trade_behavior = self_trade_behavior.DecrementTake,
        match_limit: int = None,
        use_only_deposited_funds: bool = False,
        last_valid_slot: int = None,
        last_valid_unix_timestamp: int = None,
        fail_silently_on_insufficient_funds: bool = False,
    ) -> ExecutableOrder:
        market_metadata: MarketMetadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market address not found: ", market_pubkey)

        if client_order_id is None:
            client_order_id = (
                uuid4().int
            )  # TODO: Replace with method to track orders from python client
        price_in_ticks = market_metadata.float_price_to_ticks_rounded_down(
            price_in_quote_units
        )
        num_base_lots = market_metadata.raw_base_units_to_base_lots_rounded_down(
            size_in_base_units
        )
        return ExecutableOrder(
            Limit(
                LimitValue(
                    side=side,
                    price_in_ticks=price_in_ticks,
                    num_base_lots=num_base_lots,
                    self_trade_behavior=self_trade_behavior,
                    match_limit=match_limit,
                    client_order_id=client_order_id,
                    use_only_deposited_funds=use_only_deposited_funds,
                    last_valid_slot=last_valid_slot,
                    last_valid_unix_timestamp_in_seconds=last_valid_unix_timestamp,
                    fail_silently_on_insufficient_funds=fail_silently_on_insufficient_funds,
                )
            ),
            market_pubkey=market_pubkey,
        )

    def get_post_only_order_packet(
        self,
        market_pubkey: Pubkey,
        side: SideKind,
        price_in_quote_units: float,
        size_in_base_units: float,
        client_order_id: int = None,
        reject_post_only: bool = False,
        use_only_deposited_funds: bool = False,
        last_valid_slot: int = None,
        last_valid_unix_timestamp: int = None,
        fail_silently_on_insufficient_funds: bool = False,
    ) -> ExecutableOrder:
        market_metadata: MarketMetadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market address not found: ", market_pubkey)

        if client_order_id is None:
            client_order_id = (
                uuid4().int
            )  # TODO: Replace with method to track orders from python client
        price_in_ticks = market_metadata.float_price_to_ticks_rounded_down(
            price_in_quote_units
        )
        num_base_lots = market_metadata.raw_base_units_to_base_lots_rounded_down(
            size_in_base_units
        )
        return ExecutableOrder(
            PostOnly(
                PostOnlyValue(
                    side=side,
                    price_in_ticks=price_in_ticks,
                    num_base_lots=num_base_lots,
                    client_order_id=client_order_id,
                    reject_post_only=reject_post_only,
                    use_only_deposited_funds=use_only_deposited_funds,
                    last_valid_slot=last_valid_slot,
                    last_valid_unix_timestamp_in_seconds=last_valid_unix_timestamp,
                    fail_silently_on_insufficient_funds=fail_silently_on_insufficient_funds,
                )
            ),
            market_pubkey=market_pubkey,
        )

    def get_ioc_packet(
        self,
        market_pubkey: Pubkey,
        side: SideKind,
        size_in_base_units: float,
        size_in_quote_units: float = 0,
        min_base_units_to_fill: float = 0,
        min_quote_units_to_fill: float = 0,
        price_in_quote_units: float | None = None,
        client_order_id: int = None,
        self_trade_behavior: self_trade_behavior = self_trade_behavior.DecrementTake,
        match_limit: int = None,
        use_only_deposit_funds: bool = False,
        last_valid_slot: int = None,
        last_valid_unix_timestamp: int = None,
        fail_silently_on_insufficient_funds: bool = False,
    ) -> ExecutableOrder:
        market_metadata: MarketMetadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market address not found: ", market_pubkey)

        if client_order_id is None:
            client_order_id = (
                uuid4().int
            )  # TODO: Replace with method to track orders from python client

        price_in_ticks = None
        if price_in_quote_units is not None:
            price_in_ticks = market_metadata.float_price_to_ticks_rounded_down(
                price_in_quote_units
            )

        num_base_lots = market_metadata.raw_base_units_to_base_lots_rounded_down(
            size_in_base_units
        )
        num_quote_lots = market_metadata.quote_units_to_quote_lots(size_in_quote_units)
        min_base_lots_to_fill = (
            market_metadata.raw_base_units_to_base_lots_rounded_down(
                min_base_units_to_fill
            )
        )
        min_quote_lots_to_fill = market_metadata.quote_units_to_quote_lots(
            min_quote_units_to_fill
        )

        return ExecutableOrder(
            ImmediateOrCancel(
                ImmediateOrCancelValue(
                    side=side,
                    price_in_ticks=price_in_ticks,
                    num_base_lots=num_base_lots,
                    num_quote_lots=num_quote_lots,
                    min_base_lots_to_fill=min_base_lots_to_fill,
                    min_quote_lots_to_fill=min_quote_lots_to_fill,
                    self_trade_behavior=self_trade_behavior,
                    match_limit=match_limit,
                    client_order_id=client_order_id,
                    use_only_deposited_funds=use_only_deposit_funds,
                    last_valid_slot=last_valid_slot,
                    last_valid_unix_timestamp_in_seconds=last_valid_unix_timestamp,
                    fail_silently_on_insufficient_funds=fail_silently_on_insufficient_funds,
                )
            ),
            market_pubkey=market_pubkey,
        )

    def create_place_limit_order_instruction(
        self,
        limit_order_packet: Union[Limit, PostOnly],
        market_pubkey: Pubkey,
        trader_pubkey: Pubkey,
    ) -> Instruction:
        market_metadata: MarketMetadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market not found for order: ", limit_order_packet)
        return market_metadata.create_place_limit_order_instruction(
            limit_order_packet, trader_pubkey
        )

    def create_swap_instruction(
        self,
        ioc_order_packet: ImmediateOrCancel,
        market_pubkey: Pubkey,
        trader_pubkey: Pubkey,
    ) -> Instruction:
        market_metadata: MarketMetadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market not found for order: ", ioc_order_packet)
        return market_metadata.create_swap_instruction(ioc_order_packet, trader_pubkey)

    async def send_orders(
        self,
        signer: Keypair,
        order_packets: List[ExecutableOrder],
        pre_instructions: [Instruction] = None,
        post_instructions: [Instruction] = None,
        commitment=None,
        tx_opts: TxOpts | None = None,
        recent_blockhash: Hash | None = None,
    ) -> PhoenixResponse | None:
        # Create instructions for each order_packet and create client_order_id map
        # Add instructions to transaction
        transaction = Transaction()

        # Add pre-instructions
        if pre_instructions is not None:
            for instruction in pre_instructions:
                transaction.add(instruction)

        client_orders_map: TypedDict[int, FIFOOrderId] = {}
        for executable_order in order_packets:
            order_packet = executable_order.order_packet
            market_pubkey = executable_order.market_pubkey
            if isinstance(order_packet, Limit) or isinstance(order_packet, PostOnly):
                limit_order_instruction = self.create_place_limit_order_instruction(
                    order_packet, market_pubkey, signer.pubkey()
                )
                client_orders_map[order_packet.value["client_order_id"]] = None
                transaction.add(limit_order_instruction)
            elif isinstance(order_packet, ImmediateOrCancel):
                swap_instruction = self.create_swap_instruction(
                    order_packet, market_pubkey, signer.pubkey()
                )
                transaction.add(swap_instruction)
            else:
                raise ValueError("Unrecognized order type")

        # Add pre-instructions
        if post_instructions is not None:
            for instruction in post_instructions:
                transaction.add(instruction)

        # Send transaction
        response = await self.client.send_transaction(
            transaction, signer, opts=tx_opts, recent_blockhash=recent_blockhash
        )
        signature = response.value
        # Get transaction and parse for the FIFOOrderId of each order
        commitment = commitment if commitment is not None else self.commitment
        await self.client.confirm_transaction(signature, commitment)
        transaction_response = await self.client.get_transaction(
            signature, "json", commitment, 0
        )

        transaction = transaction_response.value
        if transaction is None:
            print("Failed to confirm transaction: ", signature)
            return None

        phoenix_transaction = get_phoenix_events_from_confirmed_transaction_with_meta(
            transaction
        )
        cancelled_orders = []
        sequence_number = -1
        for ix_events in phoenix_transaction.events_from_instructions:
            sequence_number = ix_events.header.sequence_number
            for phoenix_event in ix_events.events:
                if phoenix_event.kind == "Place":
                    client_orders_map[
                        phoenix_event.value[0].client_order_id
                    ] = PhoenixOrder(
                        FIFOOrderId(
                            phoenix_event.value[0].price_in_ticks,
                            phoenix_event.value[0].order_sequence_number,
                        ),
                        phoenix_event.value[0].base_lots_placed,
                    )
                if phoenix_event.kind == "Reduce":
                    cancelled_orders.append(
                        PhoenixOrder(
                            FIFOOrderId(
                                price_in_ticks=phoenix_event.value[0].price_in_ticks,
                                order_sequence_number=phoenix_event.value[
                                    0
                                ].order_sequence_number,
                            ),
                            phoenix_event.value[0].base_lots_remaining,
                        )
                    )
        return PhoenixResponse(
            signature=signature,
            # We add 1 because the sequence number is the sequence number of the last event in the transaction
            # The current market sequence number will have been incremented
            sequence_number=sequence_number + 1,
            client_orders_map=client_orders_map,
            cancelled_orders=cancelled_orders,
        )

    def create_cancel_order_instruction(
        self,
        trader: Pubkey,
        market_pubkey: Pubkey,
        order_ids: Union[List[Union[FIFOOrderId, int]], None] = None,
        withdraw_funds=True,
    ) -> Instruction:
        log_account = Pubkey.find_program_address([b"log"], PROGRAM_ID)[0]
        market_metadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market not found: ", market_pubkey)
        if order_ids is None:
            if withdraw_funds:
                accounts = CancelAllOrdersAccounts(
                    phoenix_program=PROGRAM_ID,
                    log_authority=log_account,
                    market=market_pubkey,
                    trader=trader,
                    base_account=get_associated_token_address(
                        trader, market_metadata.base_mint
                    ),
                    quote_account=get_associated_token_address(
                        trader, market_metadata.quote_mint
                    ),
                    base_vault=market_metadata.base_vault,
                    quote_vault=market_metadata.quote_vault,
                )
                return cancel_all_orders(accounts)
            else:
                accounts = CancelAllOrdersWithFreeFundsAccounts(
                    phoenix_program=PROGRAM_ID,
                    log_authority=log_account,
                    market=market_pubkey,
                    trader=trader,
                )
                return cancel_all_orders_with_free_funds(accounts)

        else:
            order_ids = [
                FIFOOrderId.from_int(o) if isinstance(o, int) else o for o in order_ids
            ]
            orders = [
                CancelOrderParams(
                    side=Bid if (order_id.order_sequence_number & 1 << 63 > 0) else Ask,
                    price_in_ticks=order_id.price_in_ticks,
                    order_sequence_number=order_id.order_sequence_number,
                )
                for order_id in order_ids
            ]
            if withdraw_funds:
                accounts = CancelMultipleOrdersByIdAccounts(
                    phoenix_program=PROGRAM_ID,
                    log_authority=log_account,
                    market=market_pubkey,
                    trader=trader,
                    base_account=get_associated_token_address(
                        trader, market_metadata.base_mint
                    ),
                    quote_account=get_associated_token_address(
                        trader, market_metadata.quote_mint
                    ),
                    base_vault=market_metadata.base_vault,
                    quote_vault=market_metadata.quote_vault,
                )

                args = CancelMultipleOrdersByIdArgs(
                    params=CancelMultipleOrdersByIdParams(orders=orders)
                )
                return cancel_multiple_orders_by_id(args, accounts)
            else:
                accounts = CancelMultipleOrdersByIdWithFreeFundsAccounts(
                    phoenix_program=PROGRAM_ID,
                    log_authority=log_account,
                    market=market_pubkey,
                    trader=trader,
                )
                args = CancelMultipleOrdersByIdWithFreeFundsArgs(
                    params=CancelMultipleOrdersByIdParams(orders=orders)
                )
                return cancel_multiple_orders_by_id_with_free_funds(args, accounts)

    async def cancel_order(
        self,
        signer: Keypair,
        market_pubkey: Pubkey,
        order_id: Union[FIFOOrderId, int],
        **kwargs,
    ) -> Tuple[Signature, List[int]]:
        return await self.cancel_orders(
            signer=signer, market_pubkey=market_pubkey, order_ids=[order_id], **kwargs
        )

    async def cancel_all_orders(
        self,
        signer: Keypair,
        market_pubkey: Pubkey,
        **kwargs,
    ) -> Tuple[Signature, List[int]]:
        return await self.cancel_orders(
            signer=signer, market_pubkey=market_pubkey, order_ids=None, **kwargs
        )

    async def cancel_orders(
        self,
        signer: Keypair,
        market_pubkey: Pubkey,
        order_ids: Union[List[Union[FIFOOrderId, int]], None] = None,
        withdraw_cancelled_funds=True,
        withdraw_free_funds=False,
        commitment=None,
        tx_opts: TxOpts | None = None,
        recent_blockhash: Hash | None = None,
    ) -> Tuple[Signature, List[int]]:
        market_metadata = self.markets.get(market_pubkey, None)
        if market_metadata == None:
            raise ValueError("Market not found: ", market_pubkey)

        transaction = Transaction()
        transaction.add(
            self.create_cancel_order_instruction(
                signer.pubkey(),
                market_pubkey,
                order_ids,
                withdraw_funds=withdraw_cancelled_funds,
            )
        )
        if withdraw_free_funds:
            args = WithdrawFundsArgs(
                withdraw_funds_params=WithdrawParams(
                    quote_lots_to_withdraw=None,
                    base_lots_to_withdraw=None,
                )
            )
            accounts = WithdrawFundsAccounts(
                phoenix_program=PROGRAM_ID,
                log_authority=Pubkey.find_program_address([b"log"], PROGRAM_ID)[0],
                market=market_pubkey,
                trader=signer.pubkey(),
                base_account=get_associated_token_address(
                    signer.pubkey(), market_metadata.base_mint
                ),
                quote_account=get_associated_token_address(
                    signer.pubkey(), market_metadata.quote_mint
                ),
                base_vault=market_metadata.base_vault,
                quote_vault=market_metadata.quote_vault,
            )
            transaction.add(withdraw_funds(args, accounts))
        # Get transaction and parse for the FIFOOrderId of each order
        commitment = commitment if commitment is not None else self.commitment
        # Send transaction
        response = await self.client.send_transaction(
            transaction, signer, opts=tx_opts, recent_blockhash=recent_blockhash
        )
        signature = response.value
        await self.client.confirm_transaction(signature, commitment)
        transaction_response = await self.client.get_transaction(
            signature, "json", commitment, 0
        )
        transaction = transaction_response.value
        phoenix_transaction = get_phoenix_events_from_confirmed_transaction_with_meta(
            transaction
        )
        cancelled_orders: [FIFOOrderId] = []
        for ix_events in phoenix_transaction.events_from_instructions:
            for phoenix_event in ix_events.events:
                if phoenix_event.kind == "Reduce":
                    if phoenix_event.value[0].base_lots_remaining == 0:
                        cancelled_orders.append(
                            FIFOOrderId(
                                price_in_ticks=phoenix_event.value[0].price_in_ticks,
                                order_sequence_number=phoenix_event.value[
                                    0
                                ].order_sequence_number,
                            ).to_int()
                        )
        return (signature, cancelled_orders)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FIFORestingOrderJSON(typing.TypedDict):
    trader_index: int
    num_base_lots: int
    last_valid_slot: int
    last_valid_unix_timestamp_in_seconds: int


@dataclass
class FIFORestingOrder:
    layout: typing.ClassVar = borsh.CStruct(
        "trader_index" / borsh.U64,
        "num_base_lots" / borsh.U64,
        "last_valid_slot" / borsh.U64,
        "last_valid_unix_timestamp_in_seconds" / borsh.U64,
    )
    trader_index: int
    num_base_lots: int
    last_valid_slot: int
    last_valid_unix_timestamp_in_seconds: int

    @staticmethod
    def size():
        return 32

    @classmethod
    def from_decoded(cls, obj: Container) -> "FIFORestingOrder":
        return cls(
            trader_index=obj.trader_index,
            num_base_lots=obj.num_base_lots,
            last_valid_slot=obj.last_valid_slot,
            last_valid_unix_timestamp_in_seconds=obj.last_valid_unix_timestamp_in_seconds,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "trader_index": self.trader_index,
            "num_base_lots": self.num_base_lots,
            "last_valid_slot": self.last_valid_slot,
            "last_valid_unix_timestamp_in_seconds": self.last_valid_unix_timestamp_in_seconds,
        }

    def to_json(self) -> FIFORestingOrderJSON:
        return {
            "trader_index": self.trader_index,
            "num_base_lots": self.num_base_lots,
            "last_valid_slot": self.last_valid_slot,
            "last_valid_unix_timestamp_in_seconds": self.last_valid_unix_timestamp_in_seconds,
        }

    @classmethod
    def from_json(cls, obj: FIFORestingOrderJSON) -> "FIFORestingOrder":
        return cls(
            trader_index=obj["trader_index"],
            num_base_lots=obj["num_base_lots"],
            last_valid_slot=obj["last_valid_slot"],
            last_valid_unix_timestamp_in_seconds=obj[
                "last_valid_unix_timestamp_in_seconds"
            ],
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.trader_index,
                self.num_base_lots,
                self.last_valid_slot,
                self.last_valid_unix_timestamp_in_seconds,
            )
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PlaceEventJSON(typing.TypedDict):
    index: int
    order_sequence_number: int
    client_order_id: int
    price_in_ticks: int
    base_lots_placed: int


@dataclass
class PlaceEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "index" / borsh.U16,
        "order_sequence_number" / borsh.U64,
        "client_order_id" / borsh.U128,
        "price_in_ticks" / borsh.U64,
        "base_lots_placed" / borsh.U64,
    )
    index: int
    order_sequence_number: int
    client_order_id: int
    price_in_ticks: int
    base_lots_placed: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "PlaceEvent":
        return cls(
            index=obj.index,
            order_sequence_number=obj.order_sequence_number,
            client_order_id=obj.client_order_id,
            price_in_ticks=obj.price_in_ticks,
            base_lots_placed=obj.base_lots_placed,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "index": self.index,
            "order_sequence_number": self.order_sequence_number,
            "client_order_id": self.client_order_id,
            "price_in_ticks": self.price_in_ticks,
            "base_lots_placed": self.base_lots_placed,
        }

    def to_json(self) -> PlaceEventJSON:
        return {
            "index": self.index,
            "order_sequence_number": self.order_sequence_number,
            "client_order_id": self.client_order_id,
            "price_in_ticks": self.price_in_ticks,
            "base_lots_placed": self.base_lots_placed,
        }

    @classmethod
    def from_json(cls, obj: PlaceEventJSON) -> "PlaceEvent":
        return cls(
            index=obj["index"],
            order_sequence_number=obj["order_sequence_number"],
            client_order_id=obj["client_order_id"],
            price_in_ticks=obj["price_in_ticks"],
            base_lots_placed=obj["base_lots_placed"],
        )
